"""
Loader for botocore resource shapes.

Copyright 2024 Vlad Emelianov
"""

import contextlib
import importlib
from collections.abc import Iterable, Mapping
from pathlib import Path
from typing import Any, ClassVar, TypeVar, cast

from botocore.exceptions import UnknownServiceError
from botocore.loaders import Loader
from botocore.session import Session as BotocoreSession

from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.parsers.shape_parser_types import (
    PaginatorsShape,
    ResourcesShape,
    WaitersShape,
)
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.utils.boto3_utils import get_botocore_session

_T = TypeVar("_T")


class ResourceLoader:
    """
    Loader for botocore resource shapes.
    """

    SERVICE_RESOURCE_KEY = "service"
    WAITERS_KEY = "waiters-2"
    PAGINATORS_KEY = "paginators-1"
    RESOURCES_KEY = "resources-1"
    _loader: ClassVar[Loader | None] = None
    _cache: ClassVar[dict[str, Mapping[str, Any] | None]] = {}
    _botocore_session: ClassVar[BotocoreSession | None] = None

    @classmethod
    def _get_botocore_session(cls) -> BotocoreSession:
        if cls._botocore_session:
            return cls._botocore_session
        cls._botocore_session = get_botocore_session()
        return cls._botocore_session

    @classmethod
    def _get_loader(cls) -> Loader:
        if cls._loader:
            return cls._loader
        botocore_session = cls._get_botocore_session()
        loader: Loader = botocore_session.get_component("data_loader")
        cls._loader = loader
        cls._inject_boto3_path()
        return loader

    @classmethod
    def _inject_boto3_path(cls) -> None:
        """
        Backport of `boto3/session.py` `Session._setup_loader` method.
        """
        boto3_module = None
        with contextlib.suppress(ModuleNotFoundError):
            boto3_module = importlib.import_module("boto3")

        if not boto3_module:
            get_logger().debug("boto3 module not found, skipping resources injection")
            return

        if not boto3_module.__file__:
            get_logger().warning(
                "boto3 module __file__ is not defined, skipping resources injection"
            )
            return

        path = Path(boto3_module.__file__).parent / "data"
        if not cls._loader:
            raise ValueError("Loader is not initialized")
        cls._loader.search_paths.append(path.as_posix())

    def _load_resource(
        self, service_name: ServiceName, type_name: str, _response_type: type[_T]
    ) -> _T | None:
        cache_key = f"{service_name.boto3_name}.{type_name}"
        if cache_key in self._cache:
            return cast("_T | None", self._cache[cache_key])
        loader = self._get_loader()
        data: dict[str, Any] | None = None
        with contextlib.suppress(UnknownServiceError):
            data = loader.load_service_model(service_name.boto3_name, type_name)

        self._cache[cache_key] = data
        return cast("_T | None", data)

    def load_waiters(self, service_name: ServiceName) -> WaitersShape | None:
        """
        Load waiters shape.
        """
        return self._load_resource(service_name, self.WAITERS_KEY, WaitersShape)

    def load_paginators(self, service_name: ServiceName) -> PaginatorsShape | None:
        """
        Load paginators shape.
        """
        return self._load_resource(service_name, self.PAGINATORS_KEY, PaginatorsShape)

    def load_resources(self, service_name: ServiceName) -> ResourcesShape | None:
        """
        Load resources shape.
        """
        return self._load_resource(service_name, self.RESOURCES_KEY, ResourcesShape)

    def has_service_resource(self, service_name: ServiceName) -> bool:
        """
        Check if service has ServiceResource.
        """
        resources = self.load_resources(service_name)
        if not resources:
            return False
        return self.SERVICE_RESOURCE_KEY in resources

    def get_resource_service_names(self, service_names: Iterable[ServiceName]) -> list[ServiceName]:
        """
        Get service names that have a ServiceResource.
        """
        return [
            service_name
            for service_name in service_names
            if self.has_service_resource(service_name)
        ]

    @classmethod
    def get_service_data(cls, service_name: ServiceName) -> dict[str, Any]:
        """
        Get botocore service data.
        """
        botocore_session = cls._get_botocore_session()
        return botocore_session.get_service_data(service_name.boto3_name)



"""
Message-related utils.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.chat.type_defs import Message


def as_string(message: Message | list[tuple[str, str]] | str) -> str:
    """
    Convert message to string.
    """
    if isinstance(message, str):
        return message
    result = (item if isinstance(item, str) else item[1] for item in message)
    return "".join(result)


def as_message(message: Message | str) -> Message:
    """
    Convert message or string to Message.
    """
    if not message:
        return ()
    if isinstance(message, str):
        return (message,)

    return message



"""
Parsed Service package.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Generator, Iterable, Iterator, Mapping
from itertools import chain
from typing import Final, Literal

from mypy_boto3_builder.constants import TYPING_EXTENSIONS_PYPI_NAME
from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.exceptions import StructureError
from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.import_record_group import ImportRecordGroup
from mypy_boto3_builder.package_data import BasePackageData
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.class_record import ClassRecord
from mypy_boto3_builder.structures.client import Client
from mypy_boto3_builder.structures.function import Function
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.structures.package import Package
from mypy_boto3_builder.structures.paginator import Paginator
from mypy_boto3_builder.structures.service_resource import ServiceResource
from mypy_boto3_builder.structures.waiter import Waiter
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_def_sortable import TypeDefSortable
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.type_annotations.type_typed_dict import TypeTypedDict
from mypy_boto3_builder.type_annotations.type_union import TypeUnion
from mypy_boto3_builder.utils.install_requires import InstallRequiresItem
from mypy_boto3_builder.utils.strings import RESERVED_NAMES, get_anchor_link, is_reserved
from mypy_boto3_builder.utils.type_checks import is_type_def, is_typed_dict, is_union
from mypy_boto3_builder.utils.version import VersionParts, stringify_parts


class ServicePackage(Package):
    """
    Parsed Service package.
    """

    _FALLBACK_IMPORT_MAP: Final[Mapping[str, str]] = {
        Import.typing_extensions.parent: TYPING_EXTENSIONS_PYPI_NAME,
    }

    def __init__(
        self,
        *,
        data: BasePackageData,
        service_name: ServiceName,
        version: str,
        client: Client | None = None,
        service_resource: ServiceResource | None = None,
        waiters: Iterable[Waiter] = (),
        paginators: Iterable[Paginator] = (),
        type_defs: Iterable[TypeDefSortable] = (),
        literals: Iterable[TypeLiteral] = (),
        helper_functions: Iterable[Function] = (),
    ) -> None:
        super().__init__(data, (service_name,), version)
        self.pypi_name = data.get_service_pypi_name(service_name)
        self._client = client
        self.service_resource = service_resource
        self.waiters = list(waiters)
        self.paginators = list(paginators)
        self.type_defs = list(type_defs)
        self.literals = list(literals)
        self.helper_functions = list(helper_functions)
        self._annotations_import_record = ImportRecord(Import.future, "annotations")

    @property
    def name(self) -> str:
        """
        Package name.
        """
        return self.data.get_service_package_name(self.service_name)

    @property
    def client(self) -> Client:
        """
        Service Client.
        """
        if not self._client:
            raise StructureError(f"Client is not present for {self.service_name}")
        return self._client

    def extract_literals(self) -> list[TypeLiteral]:
        """
        Extract literals from children.
        """
        type_literals: set[TypeLiteral] = set()
        for type_annotation in chain(self.iterate_types(), self.type_defs):
            if isinstance(type_annotation, TypeDefSortable):
                type_literals.update(type_annotation.get_children_literals())
            if isinstance(type_annotation, TypeLiteral):
                type_literals.add(type_annotation)

        return sorted(type_literals)

    def _iterate_methods(self) -> Iterator[Method]:
        yield from self.client.methods
        for waiter in self.waiters:
            yield from waiter.methods
        for paginator in self.paginators:
            yield from paginator.methods
        if self.service_resource:
            yield from self.service_resource.methods
            for resource in self.service_resource.sub_resources:
                yield from resource.methods

    def extract_type_defs(self) -> set[TypeDefSortable]:
        """
        Extract typed defs from children.
        """
        result: set[TypeDefSortable] = set()
        for type_annotation in self.iterate_types():
            if not is_type_def(type_annotation):
                continue
            result.add(type_annotation)

        for method in self._iterate_methods():
            if not method.has_request_type_annotation():
                continue
            result.add(method.request_type_annotation)

        return result

    def get_type_defs_all_names(self) -> list[str]:
        """
        Get `__all__` statement names for `type_defs.py[i]`.
        """
        return sorted([i.name for i in self.type_defs])

    def get_literals_all_names(self) -> list[str]:
        """
        Get `__all__` statement names for `literals.py[i]`.
        """
        return sorted([i.name for i in self.literals])

    def iterate_types(self) -> Generator[FakeAnnotation]:
        """
        Iterate over type annotations from Client, ServiceResource, waiters and paginators.
        """
        yield from self.client.iterate_types()
        if self.service_resource:
            yield from self.service_resource.iterate_types()
        for waiter in self.waiters:
            yield from waiter.iterate_types()
        for paginator in self.paginators:
            yield from paginator.iterate_types()

    def get_init_import_records(self) -> ImportRecordGroup:
        """
        Get import records group for `__init__.py[i]`.
        """
        result = ImportRecordGroup()
        result.add(
            ImportRecord(
                Import.local(ServiceModuleName.client.name),
                self.client.name,
            ),
        )
        if self.service_resource:
            result.add(
                ImportRecord(
                    Import.local(ServiceModuleName.service_resource.name),
                    self.service_resource.name,
                    fallback=ImportRecord(
                        Import.builtins, "object", alias=self.service_resource.name
                    ),
                ),
            )
        for waiter in self.waiters:
            result.add(
                ImportRecord(
                    Import.local(ServiceModuleName.waiter.name),
                    waiter.name,
                ),
            )
        for paginator in self.paginators:
            result.add(
                ImportRecord(
                    Import.local(ServiceModuleName.paginator.name),
                    paginator.name,
                ),
            )

        return result

    def get_init_all_names(self) -> list[str]:
        """
        Get `__all__` statement names for `__init__.py[i]`.
        """
        names = {self.client.name, self.client.alias_name}
        if self.service_resource:
            names.update((self.service_resource.name, self.service_resource.alias_name))
        names.update(waiter.name for waiter in self.waiters)
        names.update(paginator.name for paginator in self.paginators)

        result = list(names)
        result.sort()
        return result

    def get_client_required_import_records(self) -> ImportRecordGroup:
        """
        Get import record group for `client.py[i]`.
        """
        result = ImportRecordGroup()
        result.add(self._annotations_import_record.copy())
        result.add(*self.client.get_required_import_records())
        result.add(*self.client.exceptions_class.get_required_import_records())

        return result

    def get_service_resource_required_import_records(self) -> ImportRecordGroup:
        """
        Get import record group for `service_resource.py[i]`.
        """
        result = ImportRecordGroup()
        if self.service_resource is None:
            return result

        result.add(self._annotations_import_record.copy())
        result.add(*self.service_resource.get_required_import_records())
        return result

    def get_paginator_required_import_records(self) -> ImportRecordGroup:
        """
        Get import record group for `paginator.py[i]`.
        """
        result = ImportRecordGroup()
        result.add(self._annotations_import_record.copy())
        for paginator in self.paginators:
            result.add(*paginator.get_required_import_records())

        return result

    def get_waiter_required_import_records(self) -> ImportRecordGroup:
        """
        Get import record group for `waiter.py[i]`.
        """
        result = ImportRecordGroup()
        result.add(self._annotations_import_record.copy())
        for waiter in self.waiters:
            result.add(*waiter.get_required_import_records())

        return result

    def get_type_defs_required_import_records(self) -> ImportRecordGroup:
        """
        Get import record group for `type_defs.py[i]`.
        """
        result = ImportRecordGroup()
        if not self.type_defs:
            return result

        result.add(self._annotations_import_record.copy())
        for type_def in self.type_defs:
            for import_record in type_def.get_definition_import_records():
                if import_record.source.is_type_defs():
                    continue
                result.add(import_record)

        return result

    def get_literals_required_import_records(self) -> ImportRecordGroup:
        """
        Get import record group for `literals.py[i]`.
        """
        return ImportRecordGroup(Type.Literal.get_import_records())

    def _iterate_all_class_names(self) -> Generator[str]:
        yield from (i.name for i in self.type_defs)
        yield from (i.name for i in self.literals)
        yield from (i.name for i in self.waiters)
        yield from (i.name for i in self.paginators)
        if self.service_resource:
            yield from self.service_resource.get_all_names()

    def validate(self) -> None:
        """
        Validate parsed module.

        Finds duplicated names.
        Finds conflicts with reserved Python words.
        """
        names: set[str] = set()
        for name in self._iterate_all_class_names():
            if is_reserved(name):
                raise StructureError(f"{name} is a reserved keyword")
            if name in names:
                for type_def in self.type_defs:
                    if type_def.name == name:
                        self.logger.warning(type_def.render_definition())
                raise StructureError(f"Duplicate name {name}")
            names.add(name)

    def get_doc_link(
        self,
        file: Literal[
            "client",
            "service_resource",
            "waiters",
            "paginators",
            "type_defs",
            "literals",
        ],
        *parts: str,
    ) -> str:
        """
        Get link to local docs with anchor.

        Arguments:
            file -- HTML file name
            parts -- Anchor parts
        """
        link = f"{self.get_local_doc_link()}{file}/"
        if not parts:
            return link
        anchor = "".join([get_anchor_link(part) for part in parts])
        return f"{link}#{anchor}"

    def get_local_doc_link(self) -> str:
        """
        Get link to local docs.
        """
        return self.get_service_local_doc_link(self.service_name)

    def mark_safe_typed_dicts(self) -> None:
        """
        Mark TypedDicts that can be rendered as classes safely.

        TypedDict cannot be rendered as class if its name or any attribute is a reserver word,
        or if any argument is names as another TypeDef.
        """
        unsafe_key_names = {
            *RESERVED_NAMES,
            *(type_def.name for type_def in self.type_defs),
            *(literal.name for literal in self.literals),
        }
        for type_def in self.type_defs:
            if not is_typed_dict(type_def):
                continue
            type_def.is_safe_as_class = True
            if is_reserved(type_def.name):
                type_def.is_safe_as_class = False
                continue
            if any(attribute.name in unsafe_key_names for attribute in type_def.children):
                type_def.is_safe_as_class = False
                continue

    def _get_fallback_versions(self) -> tuple[tuple[str, str], ...]:
        """
        Get max Python version that has fallbacks to typing-extensions.
        """
        result: dict[str, VersionParts] = {}
        for import_record_group in (
            self.get_literals_required_import_records(),
            self.get_client_required_import_records(),
            self.get_service_resource_required_import_records(),
            self.get_type_defs_required_import_records(),
        ):
            for import_record in import_record_group.records:
                if import_record.min_version and import_record.fallback:
                    parent = import_record.fallback.source.parent
                    result[parent] = max((result.get(parent, ()), import_record.min_version))

        return tuple((k, stringify_parts(v)) for k, v in result.items())

    def calculate_install_requires(self) -> None:
        """
        Calculate install_requires based on fallback versions.
        """
        self.install_requires.clear()
        for module_name, version in self._get_fallback_versions():
            if module_name not in self._FALLBACK_IMPORT_MAP:
                continue
            requirement_name = self._FALLBACK_IMPORT_MAP[module_name]
            self.install_requires.add(
                InstallRequiresItem(
                    requirement_name,
                    drop_python_version=version,
                )
            )

    def iterate_methods(self) -> Generator[tuple[ClassRecord, Method]]:
        """
        Iterate over all parent, method pairs.
        """
        for method in self.client.methods:
            yield self.client, method
        if self.service_resource:
            for method in self.service_resource.methods:
                yield self.service_resource, method
            for resource in self.service_resource.sub_resources:
                for method in resource.methods:
                    yield resource, method
        for waiter in self.waiters:
            for method in waiter.methods:
                yield waiter, method
        for paginator in self.paginators:
            for method in paginator.methods:
                yield paginator, method

    def iterate_typed_dicts(self, limit: int | None = None) -> Iterator[TypeTypedDict]:
        """
        Iterate over all typed dicts.
        """
        index = 0
        for type_def in self.type_defs:
            if not is_typed_dict(type_def):
                continue

            yield type_def
            index += 1
            if limit is not None and index >= limit:
                return

    def iterate_named_unions(self, limit: int | None = None) -> Iterator[TypeUnion]:
        """
        Iterate over all unions.
        """
        index = 0
        for type_def in self.type_defs:
            if not is_union(type_def):
                continue

            yield type_def
            index += 1
            if limit is not None and index >= limit:
                return



"""
GitHub-related utils.

Copyright 2024 Vlad Emelianov
"""

from io import BytesIO
from pathlib import Path
from zipfile import ZipFile

import requests

from mypy_boto3_builder.exceptions import BuildEnvError


def download_and_extract(url: str, output_path: Path) -> Path:
    """
    Download and extract zip file with stubs from GitHub URL.
    """
    response = requests.get(url, timeout=60)
    zipfile = ZipFile(BytesIO(response.content))

    if not response.ok:
        raise BuildEnvError(f"Failed to download URL {url}: {response.status_code} {response.text}")

    project_roots = [
        Path(i).parent.as_posix() for i in zipfile.namelist() if Path(i).name == "py.typed"
    ]
    if len(project_roots) != 1:
        raise BuildEnvError(f"Failed to detect project root: {project_roots}")

    project_root = project_roots[0]

    for member in zipfile.namelist():
        if not member.startswith(project_root):
            continue
        zipfile.extract(member, output_path)

    return output_path / project_root



"""
Type annotations for aiobotocore.retries.special module.

Copyright 2025 Vlad Emelianov
"""

from botocore.retries.special import RetryDDBChecksumError
from botocore.retries.standard import RetryContext

class AioRetryDDBChecksumError(RetryDDBChecksumError):
    async def is_retryable(self, context: RetryContext) -> bool: ...  # type: ignore



"""
Logging utils.

Copyright 2024 Vlad Emelianov
"""

from __future__ import annotations

import functools
import logging
import sys

import loguru

from mypy_boto3_builder.constants import LOGGER_NAME

__all__ = ("get_logger", "setup_logger")


def _formatter(name: str, record: loguru.Record) -> str:
    tags = record["extra"].get("tags") or ()
    if isinstance(tags, str):
        tags = (tags,)
    message = record["message"]
    for tag in tags:
        message = message.replace(f"{tag}", f"<cyan>{tag}</cyan>", 1)
    return (
        "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
        "<level>{level: <8}</level> | "
        f"<cyan>{name}</cyan> - "
        f"<white>{message}</white>"
        "\n"
    )


def setup_logger(level: int = logging.DEBUG, name: str = LOGGER_NAME) -> None:
    """
    Set up logger.
    """
    level_name = logging.getLevelName(level)
    loguru.logger.configure(
        handlers=[
            {"sink": sys.stderr, "level": level_name, "format": functools.partial(_formatter, name)}
        ],
    )


def get_logger() -> loguru.Logger:
    """
    Get Logger instance.
    """
    return loguru.logger


def disable_logger() -> None:
    """
    Disable logger.
    """
    loguru.logger.disable(LOGGER_NAME)


def enable_logger() -> None:
    """
    Enable logger.
    """
    loguru.logger.enable(LOGGER_NAME)



"""
Group tool for ImportRecord sets.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterable, Iterator
from typing import Final

from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.import_helpers.import_record import ImportRecord


class ImportRecordGroup:
    """
    Group tool for ImportRecord sets.
    """

    _SYS_IMPORT_RECORD: Final[ImportRecord] = ImportRecord(Import.sys)

    def __init__(self, records: Iterable[ImportRecord] = ()) -> None:
        self.records: set[ImportRecord] = set()
        self.add(*records)

    def add(self, *records: ImportRecord) -> None:
        """
        Add record to group.
        """
        for record in records:
            if record.is_implicit():
                continue
            if record.needs_sys_fallback():
                self.add(self._SYS_IMPORT_RECORD)

            self.records.add(record)

    @staticmethod
    def _render_records(records: set[ImportRecord]) -> Iterator[str]:
        sources = {x.source for x in records}
        for source in sorted(sources):
            source_records = {i for i in records if i.source == source}
            noalias_source_records = {i for i in source_records if not i.alias}
            alias_source_records = {i for i in source_records if i.alias}
            if noalias_source_records:
                names = (x.render_name() for x in sorted(noalias_source_records))
                yield f"from {source.render()} import {', '.join(names)}"
            for record in sorted(alias_source_records):
                yield f"from {source.render()} import {record.render_name()}"

    def _iterate_render_nameless(self) -> Iterator[str]:
        """
        Iterate over nameless import records.
        """
        nameless_records = {i for i in self.records if not i.name}
        for record in sorted(nameless_records):
            yield record.render()

    def _iterate_render_regular(self) -> Iterator[str]:
        """
        Iterate over rendered regular records with name and no fallback.
        """
        regular_records = {
            i for i in self.records if i.name and not i.min_version and not i.fallback
        }
        yield from self._render_records(regular_records)

    def _iterate_render_source_fallback(self) -> Iterator[str]:
        """
        Iterate over rendered records with fallback but no min version.
        """
        records = {i for i in self.records if i.fallback and not i.min_version}
        parents = {x.source.parent for x in records}
        for parent in sorted(parents):
            source_records = {i for i in records if i.source.parent == parent}
            fallback_records = {i.fallback for i in source_records if i.fallback}
            yield "\n".join(
                (
                    "try:",
                    *(f"    {x}" for x in self._render_records(source_records)),
                    "except ImportError:",
                    *(
                        f"    {x}  # type: ignore[assignment]"
                        for x in self._render_records(fallback_records)
                    ),
                ),
            )

    def _iterate_render_source_min_version(self) -> Iterator[str]:
        """
        Iterate over rendered records with min version.
        """
        records = {i for i in self.records if i.min_version}
        min_versions = {i.min_version for i in records if i.min_version}
        for min_version in sorted(min_versions):
            min_version_records = {i for i in records if i.min_version == min_version}
            min_version_str = ", ".join(str(i) for i in min_version)
            fallback_records = {i.fallback for i in min_version_records if i.fallback}
            yield "\n".join(
                (
                    f"if sys.version_info >= ({min_version_str}):",
                    *(f"    {x}" for x in self._render_records(min_version_records)),
                    "else:",
                    *(f"    {x}" for x in self._render_records(fallback_records)),
                ),
            )

    def __iter__(self) -> Iterator[str]:
        """
        Iterate over all records grouped by source.
        """
        yield from self._iterate_render_nameless()
        yield from self._iterate_render_regular()
        yield from self._iterate_render_source_fallback()
        yield from self._iterate_render_source_min_version()



"""
Type annotations for aiobotocore.endpoint module.

Copyright 2025 Vlad Emelianov
"""

from typing import Any, Mapping

from aiobotocore.awsrequest import AioAWSResponse
from aiobotocore.httpsession import AIOHTTPSession as AIOHTTPSession
from aiobotocore.response import StreamingBody as StreamingBody
from botocore.awsrequest import AWSPreparedRequest
from botocore.endpoint import DEFAULT_TIMEOUT as DEFAULT_TIMEOUT
from botocore.endpoint import MAX_POOL_CONNECTIONS as MAX_POOL_CONNECTIONS
from botocore.endpoint import Endpoint, EndpointCreator
from botocore.model import OperationModel, ServiceModel

DEFAULT_HTTP_SESSION_CLS: type[AIOHTTPSession] = ...

async def convert_to_response_dict(
    http_response: AioAWSResponse, operation_model: OperationModel
) -> dict[str, Any]: ...

class AioEndpoint(Endpoint):
    async def close(self) -> None: ...  # type: ignore[override]
    async def create_request(
        self, params: Mapping[str, Any], operation_model: OperationModel | None = ...
    ) -> AWSPreparedRequest: ...

class AioEndpointCreator(EndpointCreator):
    def create_endpoint(  # type: ignore[override]
        self,
        service_model: ServiceModel,
        region_name: str,
        endpoint_url: str,
        verify: str | bool | None = ...,
        response_parser_factory: Any | None = ...,
        timeout: int = ...,
        max_pool_connections: int = ...,
        http_session_cls: type[AIOHTTPSession] = ...,
        proxies: Any | None = ...,
        socket_options: Any | None = ...,
        client_cert: Any | None = ...,
        proxies_config: Any | None = ...,
        connector_args: Any | None = ...,
    ) -> AioEndpoint: ...



"""
Jinja2-related utils.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Mapping
from pathlib import Path
from typing import Any

from mypy_boto3_builder.jinja_manager import JinjaManager


def render_jinja2_template(template_path: Path, context: Mapping[str, Any]) -> str:
    """
    Render Jinja2 template to a string.

    Arguments:
        template_path -- Relative path to template in `TEMPLATES_PATH`
        kwargs -- Render arguments

    Returns:
        A rendered template.
    """
    template = JinjaManager().get_template(template_path)
    return template.render(context)



"""
Base wrapper for questionary.Question.

Copyright 2024 Vlad Emelianov
"""

from abc import ABC, abstractmethod
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any, ClassVar, Final, Generic, TypeVar
from unittest.mock import patch

from prompt_toolkit import PromptSession
from prompt_toolkit.key_binding.key_bindings import Binding
from prompt_toolkit.keys import Keys
from questionary.question import Question

from mypy_boto3_builder.chat.text_style import TextStyle
from mypy_boto3_builder.chat.type_defs import Message

_R = TypeVar("_R")


class BasePrompt(ABC, Generic[_R]):
    """
    Base wrapper for questionary.Question.
    """

    PROMPT_SESSION_MODULE: ClassVar = "questionary.prompts.common.PromptSession"
    STYLE: Final = TextStyle.get_style()
    HELP_MESSAGE: ClassVar[Message] = ()

    def __init__(self) -> None:
        self._question: Question | None = None
        self._result: _R | None = None

    def get_help_message(self) -> Message:
        """
        Get help message.
        """
        return self.HELP_MESSAGE

    @property
    def question(self) -> Question:
        """
        PromptSession of a currently running application.
        """
        if self._question is None:
            raise ValueError("Question is not set")
        return self._question

    def _get_enter_key_binding(self) -> Binding:
        key_bindings = self.question.application.key_bindings
        if not key_bindings:
            raise ValueError("Key bindings are not set")
        for key_binding in reversed(key_bindings.bindings):
            if key_binding.keys == (Keys.ControlM,):
                return key_binding
        raise ValueError("Key binding for Enter not found.")

    @contextmanager
    def _patch_session(self) -> Iterator[None]:
        message = self.get_prompt_tokens

        class PatchedPromptSession(PromptSession[Any]):
            """
            Patched PromptSession with custom message.
            """

            def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
                if args:
                    args = (message, *args[1:])
                super().__init__(*args, **kwargs)

        with patch(self.PROMPT_SESSION_MODULE, PatchedPromptSession):
            yield

    def get_prompt_tokens(self) -> list[tuple[str, str]]:
        """
        Get prompt tokens for PromptSession.
        """
        return TextStyle.text.stylize(self._get_prompt_tokens())

    @abstractmethod
    def _get_prompt_tokens(self) -> Message:
        raise NotImplementedError("Method is not implemented")

    def _is_answered(self) -> bool:
        return self._result is not None

    def get_question(self) -> Question:
        """
        Create an underlying question.
        """
        if self._question:
            return self._question
        with self._patch_session():
            self._question = self._create_question()
        self._patch_answer()
        return self._question

    @abstractmethod
    def _create_question(self) -> Question:
        raise NotImplementedError("Method is not implemented")

    def _patch_answer(self) -> None:
        pass

    def ask(self) -> _R:
        """
        Ask a question and return a result.
        """
        question = self.get_question()
        question.unsafe_ask()
        if self._result is None:
            raise ValueError("Result is not set")
        return self._result

    def _set_result(self, result: _R) -> None:
        """
        Set a result.
        """
        self._result = result



"""
Structure for boto3-stubs module.

Copyright 2024 Vlad Emelianov
"""

from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING

from mypy_boto3_builder.constants import PACKAGE_NAME
from mypy_boto3_builder.import_helpers.import_record_group import ImportRecordGroup
from mypy_boto3_builder.package_data import BasePackageData
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.class_record import ClassRecord
from mypy_boto3_builder.structures.package import Package
from mypy_boto3_builder.utils.install_requires import InstallRequires
from mypy_boto3_builder.utils.version import get_builder_version

if TYPE_CHECKING:
    from mypy_boto3_builder.structures.function import Function
    from mypy_boto3_builder.structures.packages.service_package import ServicePackage
    from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral


class WrapperPackage(Package, ABC):
    """
    Main package module data.
    """

    def __init__(
        self,
        data: BasePackageData,
        service_names: Iterable[ServiceName],
        version: str,
    ) -> None:
        super().__init__(data, service_names, version)
        self.session_class = ClassRecord("Session")
        self.service_packages: list[ServicePackage] = []
        self.init_functions: list[Function] = []
        self.literals: list[TypeLiteral] = []
        self.setup_package_data: dict[str, tuple[str, ...]] = (
            {self.data.name: ("py.typed", "*.pyi", "*/*.pyi")} if self.data.name else {}
        )
        self._description = self.get_short_description()
        self.install_requires = InstallRequires(self.data.install_requires)

    @property
    def description(self) -> str:
        """
        Package description for setup.py.
        """
        return self._description

    def set_description(self, value: str) -> None:
        """
        Set package description for setup.py.
        """
        self._description = value

    def get_init_required_import_records(self) -> ImportRecordGroup:
        """
        Get import record group for `__init__.py[i]`.
        """
        result = ImportRecordGroup()
        for init_function in self.init_functions:
            result.add(*init_function.get_required_import_records())

        return result

    def get_session_required_import_records(self) -> ImportRecordGroup:
        """
        Get import record group for `session.py[i]`.
        """
        return ImportRecordGroup(self.session_class.get_required_import_records())

    @abstractmethod
    def get_all_names(self) -> list[str]:
        """
        Get names for `__all__` directive.
        """
        raise NotImplementedError("Method should be implemented in child class")

    def get_short_description(self, prefix: str = "Type annotations") -> str:
        """
        Get short description for the package in setup.py.
        """
        return (
            f"{prefix} for {self.library_name} {self.library_version}"
            f" generated with {PACKAGE_NAME} {get_builder_version()}"
        )



"""
Type annotations for aiobotocore.client module.

Copyright 2025 Vlad Emelianov
"""

from types import TracebackType
from typing import Any, TypeVar

from aiobotocore.paginate import AioPaginator
from aiobotocore.waiter import AIOWaiter
from botocore.client import BaseClient, ClientCreator
from botocore.config import Config
from botocore.history import HistoryRecorder

_R = TypeVar("_R")

history_recorder: HistoryRecorder

class AioClientCreator(ClientCreator):
    async def create_client(  # type: ignore[override]
        self,
        service_name: str,
        region_name: str,
        is_secure: bool = ...,
        endpoint_url: str | None = ...,
        verify: str | bool | None = ...,
        credentials: Any | None = ...,
        scoped_config: Any | None = ...,
        api_version: str | None = ...,
        client_config: Config | None = ...,
        auth_token: str | None = ...,
    ) -> AioBaseClient: ...

class AioBaseClient(BaseClient):
    def get_paginator(self, operation_name: str) -> AioPaginator[Any]: ...
    def get_waiter(self, waiter_name: str) -> AIOWaiter: ...
    async def __aenter__(self: _R) -> _R: ...
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        tb: TracebackType | None,
    ) -> None: ...
    async def close(self) -> None: ...  # type: ignore[override]



"""
Predefined FakeAnnotation instances.

Copyright 2024 Vlad Emelianov
"""

from datetime import datetime
from decimal import Decimal
from typing import Final

from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_annotation import TypeAnnotation
from mypy_boto3_builder.type_annotations.type_constant import TypeConstant
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript


class Type:
    """
    Predefined FakeAnnotation instances.
    """

    Union: Final[TypeAnnotation] = TypeAnnotation("Union")
    Any: Final[TypeAnnotation] = TypeAnnotation("Any")
    Dict: Final[TypeAnnotation] = TypeAnnotation("Dict")
    List: Final[TypeAnnotation] = TypeAnnotation("List")
    Set: Final[TypeAnnotation] = TypeAnnotation("Set")
    Mapping: Final[TypeAnnotation] = TypeAnnotation("Mapping")
    Sequence: Final[TypeAnnotation] = TypeAnnotation("Sequence")
    Callable: Final[TypeAnnotation] = TypeAnnotation("Callable")
    Awaitable: Final[TypeAnnotation] = TypeAnnotation("Awaitable")
    TypedDict: Final[TypeAnnotation] = TypeAnnotation("TypedDict")
    Literal: Final[TypeAnnotation] = TypeAnnotation("Literal")
    IO: Final[TypeAnnotation] = TypeAnnotation("IO")
    overload: Final[TypeAnnotation] = TypeAnnotation("overload")
    none: Final[TypeConstant] = TypeConstant(None)
    str: Final[ExternalImport] = ExternalImport.from_class(str)
    bool: Final[ExternalImport] = ExternalImport.from_class(bool)
    bytes: Final[ExternalImport] = ExternalImport.from_class(bytes)
    bytearray: Final[ExternalImport] = ExternalImport.from_class(bytearray)
    int: Final[ExternalImport] = ExternalImport.from_class(int)
    float: Final[ExternalImport] = ExternalImport.from_class(float)
    Ellipsis: Final[TypeConstant] = TypeConstant(TypeConstant.Ellipsis)
    Decimal: Final[ExternalImport] = ExternalImport.from_class(Decimal)
    Type: Final[TypeAnnotation] = TypeAnnotation("Type")
    Iterator: Final[TypeAnnotation] = TypeAnnotation("Iterator")
    AsyncIterator: Final[TypeAnnotation] = TypeAnnotation("AsyncIterator")
    datetime: Final[ExternalImport] = ExternalImport.from_class(datetime)

    SequenceAny: Final[TypeSubscript] = TypeSubscript(Sequence, [Any])
    MappingStrAny: Final[TypeSubscript] = TypeSubscript(Mapping, [str, Any])
    ListAny: Final[TypeSubscript] = TypeSubscript(List, [Any])
    DictStrAny: Final[TypeSubscript] = TypeSubscript(Dict, [str, Any])
    DictStrStr: Final[TypeSubscript] = TypeSubscript(Dict, [str, str])
    IOAny: Final[TypeSubscript] = TypeSubscript(IO, [Any])
    RemoveArgument: Final[TypeConstant] = TypeConstant("RemoveArgument")
    NotRequired: Final[TypeAnnotation] = TypeAnnotation("NotRequired")
    NoReturn: Final[TypeAnnotation] = TypeAnnotation("NoReturn")
    Unpack: Final[TypeAnnotation] = TypeAnnotation("Unpack")
    Self: Final[TypeAnnotation] = TypeAnnotation("Self")

    @classmethod
    def unpack(cls, wrapped: FakeAnnotation) -> FakeAnnotation:
        """
        Get Unpack type annotation.
        """
        return TypeSubscript(cls.Unpack, [wrapped])

    @classmethod
    def wrap_list(cls, wrapped: FakeAnnotation) -> FakeAnnotation:
        """
        Get List type annotation.
        """
        return TypeSubscript(cls.List, [wrapped])



"""
Parser for boto3 changelog.

Copyright 2024 Vlad Emelianov
"""

import re

import requests

from mypy_boto3_builder.constants import REQUEST_TIMEOUT
from mypy_boto3_builder.exceptions import BuildEnvError


class BotocoreChangelog:
    """
    Parser for boto3 changelog.
    """

    URL = "https://raw.githubusercontent.com/boto/botocore/develop/CHANGELOG.rst"
    SERVICE_NAME_RE = re.compile(r"``(\S+)``")

    def __init__(self) -> None:
        self._changelog: str = ""

    @classmethod
    def _get_changelog(cls) -> str:
        response = requests.get(cls.URL, timeout=REQUEST_TIMEOUT)
        if not response.ok:
            raise BuildEnvError(
                f"Cannot retrieve {cls.URL}: {response.status_code} {response.text}",
            ) from None

        return response.text

    def _get_section(self, version: str) -> str:
        result: list[str] = []
        self._changelog = self._changelog or self._get_changelog()
        found = False
        for line in self._changelog.splitlines():
            if line == version:
                found = True
                continue
            if found and result and line.startswith("==="):
                result.pop()
                break
            if found:
                result.append(line)

        return "\n".join(result[1:])

    def fetch_updated(self, version: str) -> list[str]:
        """
        Get a list of service names updated in `version` release.
        """
        result: list[str] = []
        section = self._get_section(version)
        for match in self.SERVICE_NAME_RE.findall(section):
            if match not in result:
                result.append(match)

        return result



"""
Package links structure.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.package_data import BasePackageData
from mypy_boto3_builder.utils.strings import get_pypi_link


class PackageURL:
    """
    Package links structure.
    """

    def __init__(self, pypi_name: str, data: BasePackageData) -> None:
        self.pypi_name = pypi_name
        self.data = data

    @property
    def pypi_badge(self) -> str:
        """
        Package name badge from shields.io.
        """
        return f"https://img.shields.io/pypi/v/{self.pypi_name}.svg?color=blue"

    @property
    def pyversions_badge(self) -> str:
        """
        Package Python version badge from shields.io.
        """
        return f"https://img.shields.io/pypi/pyversions/{self.pypi_name}.svg?color=blue"

    @property
    def rtd_badge(self) -> str:
        """
        ReadTheDocs badge from shields.io.
        """
        return "https://img.shields.io/readthedocs/boto3-stubs.svg?color=blue"

    @property
    def pepy_badge(self) -> str:
        """
        PePy total downloads badge.
        """
        return f"https://static.pepy.tech/badge/{self.pypi_name}"

    @property
    def montly_downloads_badge(self) -> str:
        """
        PyPi monthly downloads badge.
        """
        return f"https://img.shields.io/pypi/dm/{self.pypi_name}?color=blue"

    @property
    def pypi(self) -> str:
        """
        Get link to PyPI.
        """
        return get_pypi_link(self.pypi_name)

    @property
    def library_pypi(self) -> str:
        """
        Get link to PyPI underlying library.
        """
        return get_pypi_link(self.data.get_library_name())

    @property
    def stubs_pypi(self) -> str:
        """
        Get link to PyPI lite version of stubs.
        """
        return get_pypi_link(self.data.pypi_stubs_name)

    @property
    def stubs_lite_pypi(self) -> str:
        """
        Get link to PyPI lite version of stubs.
        """
        return get_pypi_link(self.data.pypi_lite_name)

    @property
    def stubs_full_pypi(self) -> str:
        """
        Get link to PyPI full version of stubs.
        """
        return get_pypi_link(self.data.pypi_full_name)

    @property
    def pepy(self) -> str:
        """
        PePy project link.
        """
        return f"https://pepy.tech/project/{self.pypi_name}"

    @property
    def pypistats(self) -> str:
        """
        PePy project link.
        """
        return f"https://pypistats.org/packages/{self.pypi_name}"

    @property
    def docs(self) -> str:
        """
        Documentation link.
        """
        return self.data.local_doc_link



"""
RDS service injected methods.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.utils.type_checks import get_optional

generate_db_auth_token_method = Method(
    name="generate_db_auth_token",
    arguments=(
        Argument.self(),
        Argument("DBHostname", Type.str),
        Argument("Port", Type.int),
        Argument("DBUsername", Type.str),
        Argument("Region", get_optional(Type.str), Type.Ellipsis),
    ),
    return_type=Type.str,
    docstring="Generates an auth token used to connect to a db with IAM credentials.",
)

CLIENT_METHODS = (generate_db_auth_token_method.copy(),)



from pathlib import Path

import pytest

from mypy_boto3_builder.exceptions import JinjaManagerError
from mypy_boto3_builder.jinja_manager import JinjaManager


class TestJinjaManager:
    def test_init(self) -> None:
        assert JinjaManager()

    def test_get_template(self) -> None:
        manager = JinjaManager()
        assert manager.get_template(Path("common/named_union.py.jinja2"))
        assert manager.get_template(Path("common/named_union.py.jinja2"))
        with pytest.raises(JinjaManagerError):
            manager.get_template(Path("common/unknown.jinja2"))



"""
Enum with all parent imports.

Copyright 2024 Vlad Emelianov
"""

from enum import Enum
from typing import Final


class ImportParent(Enum):
    """
    Enum with all parent imports.
    """

    future = "__future__"
    builtins = "builtins"
    boto3 = "boto3"
    botocore = "botocore"
    typing = "typing"
    awscrt = "awscrt"
    s3transfer = "s3transfer"
    aiobotocore = "aiobotocore"
    aioboto3 = "aioboto3"
    typing_extensions = "typing_extensions"
    types = "types"
    sys = "sys"
    collections = "collections"

    @classmethod
    def is_third_party(cls, parent_name: str) -> bool:
        """
        Whether import is from 3rd party module.
        """
        return parent_name in _THIRD_PARTY

    @classmethod
    def is_builtins(cls, parent_name: str) -> bool:
        """
        Whether import is from Python `builtins` module.
        """
        return parent_name == cls.builtins.value


_THIRD_PARTY: Final = frozenset(
    (
        ImportParent.boto3.value,
        ImportParent.botocore.value,
        ImportParent.aioboto3.value,
        ImportParent.aiobotocore.value,
        ImportParent.s3transfer.value,
        ImportParent.awscrt.value,
    )
)



"""
String to type annotation map to replace overriden botocore shapes.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Mapping
from typing import Final

from mypy_boto3_builder.constants import ALL, SERVICE_RESOURCE
from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_maps.named_unions import (
    AttributeValueTypeDef,
    BlobTypeDef,
    PolicyDocumentTypeDef,
    StreamingBodyType,
    TableAttributeValueTypeDef,
    TimestampTypeDef,
    UniversalAttributeValueTypeDef,
)
from mypy_boto3_builder.type_maps.typed_dicts import GetTemplateOutputTypeDef
from mypy_boto3_builder.utils.lookup_dict import LookupDict

ShapeTypeMap = Mapping[ServiceName, Mapping[str, Mapping[str, FakeAnnotation]]]

# Mapping overriding TypeDefs for botocore shapes.
# ServiceName -> TypedDict name -> Shape name -> Type annotation
# Missing value is parsed from botocore shapes.
SHAPE_TYPE_MAP: Final[ShapeTypeMap] = {
    ServiceNameCatalog.all: {
        ALL: {
            "integer": Type.int,
            "long": Type.int,
            "boolean": Type.bool,
            "double": Type.float,
            "float": Type.float,
            "timestamp": TimestampTypeDef,
            "blob": BlobTypeDef,
            "blob_streaming": BlobTypeDef,
        },
    },
    ServiceNameCatalog.dynamodb: {
        ALL: {
            "AttributeValueTypeDef": UniversalAttributeValueTypeDef,
            "ConditionExpressionTypeDef": Type.bool,
        },
        SERVICE_RESOURCE: {
            "AttributeValueTypeDef": TableAttributeValueTypeDef,
        },
        "Table": {
            "AttributeValueTypeDef": TableAttributeValueTypeDef,
        },
    },
}

_SHAPE_TYPE_MAP_LOOKUP: LookupDict[FakeAnnotation] = LookupDict(
    {ServiceNameCatalog.to_str(k): v for k, v in SHAPE_TYPE_MAP.items()},
)

OUTPUT_SHAPE_TYPE_MAP: ShapeTypeMap = {
    ServiceNameCatalog.all: {
        ALL: {
            "timestamp": Type.datetime,
            "blob": Type.bytes,
            "blob_streaming": StreamingBodyType,
        },
    },
    ServiceNameCatalog.iam: {
        ALL: {
            "policyDocumentType": PolicyDocumentTypeDef,
        },
    },
    ServiceNameCatalog.dynamodb: {
        ALL: {
            "AttributeValueTypeDef": AttributeValueTypeDef,
        },
        SERVICE_RESOURCE: {
            "AttributeValueTypeDef": TableAttributeValueTypeDef,
        },
        "Table": {
            "AttributeValueTypeDef": TableAttributeValueTypeDef,
        },
    },
    # FIXME: botocore processes TemplateBody with json_decode_template_body
    ServiceNameCatalog.cloudformation: {
        ALL: {
            "GetTemplateOutputTypeDef": GetTemplateOutputTypeDef,
        },
    },
}

_OUTPUT_SHAPE_TYPE_MAP_LOOKUP: LookupDict[FakeAnnotation] = LookupDict(
    {ServiceNameCatalog.to_str(k): v for k, v in OUTPUT_SHAPE_TYPE_MAP.items()},
)


def get_shape_type_stub(
    service_name: ServiceName,
    resource_name: str,
    shape_name: str,
) -> FakeAnnotation | None:
    """
    Get stub type for input botocore shape.

    Arguments:
        service_name -- Service name
        resource_name -- Resource name
        shape_name -- Target Shape name

    Returns:
        Type annotation or None.
    """
    return _SHAPE_TYPE_MAP_LOOKUP.get(service_name.name, resource_name, shape_name)


def get_output_shape_type_stub(
    service_name: ServiceName,
    resource_name: str,
    shape_name: str,
) -> FakeAnnotation | None:
    """
    Get stub type for output botocore shape.

    Arguments:
        shape_type_map -- Map to lookup
        service_name -- Service name
        resource_name -- Resource name
        shape_name -- Target Shape name

    Returns:
        Type annotation or None.
    """
    type_annotation = _OUTPUT_SHAPE_TYPE_MAP_LOOKUP.get(
        service_name.name,
        resource_name,
        shape_name,
    )
    if type_annotation:
        return type_annotation

    return get_shape_type_stub(service_name, resource_name, shape_name)



"""
Fake parser that produces `structures.ServiceModule` for main module and stubs.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.package_data import BasePackageData
from mypy_boto3_builder.parsers.shape_parser import ShapeParser
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.client import Client
from mypy_boto3_builder.structures.packages.service_package import ServicePackage
from mypy_boto3_builder.structures.paginator import Paginator
from mypy_boto3_builder.structures.service_resource import ServiceResource
from mypy_boto3_builder.structures.waiter import Waiter
from mypy_boto3_builder.utils.strings import xform_name


def parse_fake_service_package(
    service_name: ServiceName,
    package_data: BasePackageData,
    version: str,
) -> ServicePackage:
    """
    Create fake boto3 service module structure.

    Used by stubs and main package.

    Arguments:
        service_name -- Target service name.
        package_data -- Package data.
        version -- Package version.

    Returns:
        ServiceModule structure.
    """
    shape_parser = ShapeParser(service_name)

    result = ServicePackage(
        data=package_data,
        service_name=service_name,
        client=Client(
            name=service_name.get_client_name(),
            service_name=service_name,
        ),
        version=version,
    )

    if shape_parser.has_service_resource():
        result.service_resource = ServiceResource(
            name=service_name.get_service_resource_name(),
            service_name=service_name,
        )

    for waiter_name in shape_parser.get_waiter_names():
        waiter = Waiter(
            name=f"{waiter_name}Waiter",
            waiter_name=waiter_name,
            attribute_name=xform_name(waiter_name),
            service_name=service_name,
        )
        result.waiters.append(waiter)

    for paginator_name in shape_parser.get_paginator_names():
        operation_name = xform_name(paginator_name)
        result.paginators.append(
            Paginator(
                f"{paginator_name}Paginator",
                operation_name=operation_name,
                service_name=service_name,
                paginator_name=paginator_name,
            ),
        )

    return result



"""
Parent class for all type annotation wrappers.

Copyright 2024 Vlad Emelianov
"""

import copy
import functools
from abc import ABC, abstractmethod
from collections.abc import Generator
from typing import Self

from mypy_boto3_builder.exceptions import BuildInternalError
from mypy_boto3_builder.import_helpers.import_record import ImportRecord


@functools.total_ordering
class FakeAnnotation(ABC):
    """
    Parent class for all type annotation wrappers.
    """

    def __hash__(self) -> int:
        """
        Calculate hash value based on string render.
        """
        return hash(self.get_sort_key())

    def __eq__(self, other: object) -> bool:
        """
        Whether two annotations are equal.
        """
        if not isinstance(other, FakeAnnotation):
            raise BuildInternalError(f"{other} is not FakeAnnotation")

        return self.get_sort_key() == other.get_sort_key()

    def __gt__(self, other: "FakeAnnotation") -> bool:
        """
        Compare two annotations for sorting.
        """
        return self.get_sort_key() > other.get_sort_key()

    def get_sort_key(self) -> str:
        """
        Get string to sort annotations.
        """
        return str(self)

    def __str__(self) -> str:
        """
        Render annotation usage as a valid Python statement.
        """
        return self.render()

    @abstractmethod
    def render(self) -> str:
        """
        Render type annotation to a valid Python code for local usage.
        """

    def _get_import_records(self) -> set[ImportRecord]:
        """
        Get import record required for using type annotation.
        """
        return set()

    def get_import_records(self) -> set[ImportRecord]:
        """
        Get all import records required for using type annotation.
        """
        return {
            import_record
            for import_record in self._get_import_records()
            if not import_record.is_implicit()
        }

    def iterate_types(self) -> Generator["FakeAnnotation"]:
        """
        Iterate over all used type annotations recursively including self.
        """
        yield self

    def is_dict(self) -> bool:
        """
        Whether type annotation is `Dict` or `TypedDict`.
        """
        return False

    def is_list(self) -> bool:
        """
        Whether type annotation is `List`.
        """
        return False

    def is_literal(self) -> bool:
        """
        Whether type annotation is `Literal`.
        """
        return False

    @abstractmethod
    def __copy__(self) -> Self:
        """
        Create a copy of type annotation wrapper.
        """

    def copy(self) -> Self:
        """
        Create a copy of type annotation wrapper.
        """
        return copy.copy(self)

    def get_local_types(self) -> list["FakeAnnotation"]:
        """
        Get internal types generated by builder.
        """
        return []

    def render_definition(self) -> str:
        """
        Render type annotation for debug purposes.
        """
        return self.render()



"""
Select multiple prompt based on questionary.path.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Sequence

import questionary
from prompt_toolkit.key_binding.key_processor import KeyPressEvent

from mypy_boto3_builder.chat.choice import Choice
from mypy_boto3_builder.chat.prompts.select_prompt import SelectPrompt
from mypy_boto3_builder.chat.text_style import TextStyle
from mypy_boto3_builder.chat.type_defs import Message
from mypy_boto3_builder.chat.utils import as_string


class MultiSelectPrompt(SelectPrompt):
    """
    Select multiple prompt based on questionary.path.
    """

    HELP_MESSAGE = (
        "Select items using ",
        TextStyle.tag.wrap("arrow keys"),
        ", press ",
        TextStyle.tag.wrap("Enter"),
        " to add or remove item.",
    )

    def __init__(
        self,
        message: Message | str,
        choices: Sequence[Choice | str],
        *,
        default: Choice | str | None = None,
        message_end: Message | str = "",
        instruction: Message | str = "",
        finish_choice: Choice | str = "Go back",
        finish_selected_text: Message | str | None = None,
    ) -> None:
        self.finish_choice = Choice.create(finish_choice)
        self.finish_text = self.finish_choice.text
        self.finish_selected_text = (
            finish_selected_text if finish_selected_text is not None else self.finish_text
        )
        self._selected_choices: list[Choice] = []
        super().__init__(
            message=message,
            choices=(self.finish_choice, *choices),
            default=default,
            message_end=message_end,
            instruction=instruction,
        )
        self._selectable_choices = tuple(
            i for i in self.choices if i != self.finish_choice and not i.disabled
        )

    def get_help_message(self) -> Message:
        """
        Get help message.
        """
        return (
            *super().get_help_message(),
            " Select ",
            TextStyle.tag.wrap(as_string(self.finish_choice.raw_title)),
            " to continue.",
        )

    def _is_all_selected(self) -> bool:
        return len(self._selected_choices) == len(self._selectable_choices)

    def _update_result(self) -> None:
        selected_choice = self._get_pointed_at()
        if selected_choice == self.finish_choice:
            self._set_result(self._selected_choices)
            return
        if selected_choice in self._selected_choices:
            self._selected_choices.remove(selected_choice)
            selected_choice.deselect()
        else:
            self._selected_choices.append(selected_choice)
            select_index = max(i.select_index for i in self.choices) + 1
            selected_choice.select(select_index)
        self.finish_choice.text = (
            self.finish_selected_text if self._selected_choices else self.finish_text
        )
        if self._is_all_selected():
            self._set_result(self._selected_choices)

    def _patch_answer(self) -> None:
        enter_key_binding = self._get_enter_key_binding()

        def set_answer(event: KeyPressEvent) -> None:
            self._update_result()
            if self._result is not None:
                self.inquirer_control.is_answered = True
                event.app.exit(result=self._result)

        enter_key_binding.handler = set_answer

    def _create_question(self) -> questionary.Question:
        """
        Create an underlying question.
        """
        return questionary.select(
            message="",
            choices=self.choices,
            default=self.default,
            style=self.STYLE,
            use_shortcuts=True,
            use_jk_keys=False,
        )



"""
Boto3 ServiceResource collections parser, produces `structures.Collection`.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.boto3_ports.model import Collection as Boto3Collection
from mypy_boto3_builder.boto3_ports.model import ResourceModel
from mypy_boto3_builder.parsers.shape_parser import ShapeParser
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.collection import Collection
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.utils.strings import get_class_prefix


def parse_collection(
    collection: Boto3Collection,
    service_name: ServiceName,
    resource_name: str,
    shape_parser: ShapeParser,
) -> Collection:
    """
    Parse Collection from boto3 resource.

    Arguments:
        collection -- Boto3 collection.
        service_name -- Service name.
        resource_name -- Resource name.
        shape_parser -- Shape parser.
    """
    if not collection.resource:
        raise ValueError(f"Collection {collection.name} does not have resource")
    object_class_name = collection.resource.type
    collection_record = Collection(
        name=f"{resource_name}{get_class_prefix(collection.name)}Collection",
        parent_name=resource_name,
        attribute_name=collection.name,
        service_name=service_name,
        type_annotation=InternalImport(collection.name),
        object_class_name=object_class_name,
    )
    self_type = InternalImport(collection_record.name)

    # FIXME: other collection methods use different anchor format. This will probably change.
    all_method = Method(
        name="all",
        arguments=(Argument.self(),),
        return_type=self_type,
        docstring=(
            "Get all items from the collection, optionally"
            " with a custom page size and item count limit."
        ),
        boto3_doc_link=f"{collection_record.boto3_doc_link_parent}#{service_name.class_name}.{collection_record.parent_name}.all",
    )

    collection_record.methods.append(all_method)
    filter_method = shape_parser.get_collection_filter_method(
        collection_record.name,
        collection,
        self_type,
    )
    filter_method.docstring = (
        "Get items from the collection, passing keyword arguments along"
        " as parameters to the underlying service operation, which are"
        " typically used to filter the results."
    )
    filter_method.type_ignore = "override"
    collection_record.methods.append(filter_method)
    batch_methods = shape_parser.get_collection_batch_methods(collection_record.name, collection)
    for batch_method in batch_methods:
        batch_method.docstring = "Batch method."
        collection_record.methods.append(batch_method)

    extra_methods = (
        Method(
            name="limit",
            arguments=[Argument.self(), Argument("count", Type.int)],
            return_type=self_type,
            docstring=f"Return at most this many {object_class_name}s.",
        ),
        Method(
            name="page_size",
            arguments=[Argument.self(), Argument("count", Type.int)],
            return_type=self_type,
            docstring=f"Fetch at most this many {object_class_name}s per service request.",
        ),
        Method(
            name="pages",
            arguments=[Argument.self()],
            return_type=TypeSubscript(
                Type.Iterator,
                [Type.wrap_list(InternalImport(name=object_class_name))],
            ),
            docstring=f"A generator which yields pages of {object_class_name}s.",
        ),
        Method(
            name="__iter__",
            arguments=[Argument.self()],
            return_type=TypeSubscript(Type.Iterator, [InternalImport(name=object_class_name)]),
            docstring=f"A generator which yields {object_class_name}s.",
        ),
    )
    collection_record.methods.extend(extra_methods)
    return collection_record


def parse_collections(
    service_name: ServiceName,
    resource_name: str,
    resource_model: ResourceModel,
    shape_parser: ShapeParser,
) -> list[Collection]:
    """
    Extract collections from boto3 resource.

    Arguments:
        service_name -- Service name.
        resource_name -- Resource name.
        resource -- boto3 service resource.
        shape_parser -- Shape parser.
    """
    result: list[Collection] = []
    for collection in resource_model.collections:
        if not collection.resource:
            continue
        collection_record = parse_collection(collection, service_name, resource_name, shape_parser)
        result.append(collection_record)
    return result



"""
Postprocessor for aiobotocore classes and methods.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Generator, Iterable
from typing import Final

from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.postprocessors.aio_imports import replace_import_with_aio
from mypy_boto3_builder.postprocessors.base import BasePostprocessor
from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.collection import Collection
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_maps.aio_resource_method_map import get_aio_resource_method
from mypy_boto3_builder.utils.type_checks import get_optional, is_external_import, is_type_parent


class AioBotocorePostprocessor(BasePostprocessor):
    """
    Postprocessor for aiobotocore classes and methods.
    """

    _COMMON_COLLECTION_METHOD_NAMES: Final = {
        "__iter__",
        "__aiter__",
        "all",
        "pages",
        "filter",
        "limit",
        "page_size",
    }
    _NOT_ASYNC_METHOD_NAMES: Final = {
        "exceptions",
        "get_waiter",
        "get_paginator",
        "can_paginate",
    }

    def process_package(self) -> None:
        """
        Convert all methods to asynchronous.
        """
        self._make_async_client()
        self._make_async_waiters()
        self._make_async_service_resource()
        self._make_async_collections()
        self._make_async_sub_resources()
        self._add_contextmanager_methods()
        self._replace_botocore_external_imports()

    def _make_async_client(self) -> None:
        for method in self.package.client.methods:
            if method.name in self._NOT_ASYNC_METHOD_NAMES:
                continue
            method.is_async = True

    @classmethod
    def _make_async_collection(cls, collection: Collection) -> None:
        for method in collection.methods:
            if method.name not in cls._COMMON_COLLECTION_METHOD_NAMES:
                method.is_async = True

        pages_method = collection.get_method("pages")
        if not isinstance(pages_method.return_type, TypeSubscript):
            raise TypeError(
                f"{collection.name}.pages method return type is not TypeSubscript:"
                f" {pages_method.return_type.render()}",
            )
        pages_method.return_type.parent = Type.AsyncIterator
        pages_method.type_ignore = "override"

        aiter_method = collection.get_method("__iter__").copy()
        aiter_method.name = "__aiter__"
        if not isinstance(aiter_method.return_type, TypeSubscript):
            raise TypeError(
                f"{collection.name}.__aiter__ method return type is not TypeSubscript:"
                f" {aiter_method.return_type.render()}",
            )
        aiter_method.return_type.parent = Type.AsyncIterator
        collection.methods.append(aiter_method)

        iter_method = collection.get_method("__iter__")
        iter_method.return_type = Type.NoReturn

    def _make_async_collections(self) -> None:
        if not self.package.service_resource:
            return
        for collection in self.package.service_resource.collections:
            self._make_async_collection(collection)

    def _make_async_waiters(self) -> None:
        for waiter in self.package.waiters:
            for method in waiter.methods:
                method.is_async = True

    def _make_async_service_resource(self) -> None:
        if not self.package.service_resource:
            return
        for method in self.package.service_resource.methods:
            method.is_async = True

    def _make_async_sub_resources(self) -> None:
        if not self.package.service_resource:
            return
        for sub_resource in self.package.service_resource.sub_resources:
            for method in sub_resource.methods:
                method_stub = get_aio_resource_method(
                    self.package.service_name,
                    sub_resource.name,
                    method.name,
                )
                if method_stub:
                    method.arguments = method_stub.arguments
                    method.return_type = method_stub.return_type
                    method.is_async = method_stub.is_async
                    continue
                method.is_async = True
            for collection in sub_resource.collections:
                self._make_async_collection(collection)
            for attribute in sub_resource.attributes:
                if not attribute.is_autoload_property():
                    continue
                attribute.type_annotation = TypeSubscript(
                    Type.Awaitable,
                    [attribute.type_annotation],
                )

    def _add_contextmanager_methods(self) -> None:
        self.package.client.methods.extend(
            (
                Method(
                    name="__aenter__",
                    arguments=(Argument.self(),),
                    return_type=Type.Self,
                    is_async=True,
                    docstring=self.package.client.docstring,
                    boto3_doc_link=self.package.client.boto3_doc_link,
                ),
                Method(
                    name="__aexit__",
                    arguments=(
                        Argument.self(),
                        Argument(
                            "exc_type",
                            get_optional(
                                TypeSubscript(Type.Type, [ExternalImport.from_class(BaseException)])
                            ),
                        ),
                        Argument("exc_val", get_optional(ExternalImport.from_class(BaseException))),
                        Argument(
                            "exc_tb",
                            get_optional(ExternalImport(Import.types, "TracebackType")),
                        ),
                    ),
                    return_type=Type.none,
                    is_async=True,
                    docstring=self.package.client.docstring,
                    boto3_doc_link=self.package.client.boto3_doc_link,
                ),
            ),
        )

    def _iterate_types_shallow(self) -> Generator[FakeAnnotation]:
        yield from self.package.client.iterate_types()
        for paginator in self.package.paginators:
            yield from paginator.iterate_types()
        for waiter in self.package.waiters:
            yield from waiter.iterate_types()
        if self.package.service_resource:
            yield from self.package.service_resource.iterate_types()

    @classmethod
    def _iterate_external_imports(
        cls,
        type_annotations: Iterable[FakeAnnotation],
    ) -> Generator[ExternalImport]:
        stack = list(type_annotations)
        while stack:
            type_annotation = stack.pop()
            if is_external_import(type_annotation):
                yield type_annotation
            if is_type_parent(type_annotation):
                stack.extend(type_annotation.get_children_types())

    def _replace_botocore_external_imports(self) -> None:
        shallow_type_annotations = self._iterate_types_shallow()
        for external_import in self._iterate_external_imports(shallow_type_annotations):
            replace_import_with_aio(external_import)
            if external_import.source.startswith(Import.boto3):
                external_import.fallback = ImportRecord(
                    Import.builtins, "object", alias=external_import.name
                )



"""
Package extra entry.

Copyright 2024 Vlad Emelianov
"""

from typing import NamedTuple


class PackageExtra(NamedTuple):
    """
    Package extra entry.
    """

    name: str
    packages: tuple[str, ...]
    description: str = ""



"""
Type definitions for builder.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Sequence
from typing import TypedDict

from mypy_boto3_builder.cli_parser import CLINamespace
from mypy_boto3_builder.enums.product import Product
from mypy_boto3_builder.service_name import ServiceName


class GeneratorKwargs(TypedDict):
    """
    Generator keyword arguments.
    """

    product: Product
    service_names: Sequence[ServiceName]
    main_service_names: Sequence[ServiceName]
    config: CLINamespace
    version: str
    cleanup: bool



"""
EC2 service injected methods.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_maps.typed_dicts import TagTypeDef

create_tags_method = Method(
    name="create_tags",
    arguments=(
        Argument.self(),
        Argument.kwflag(),
        Argument("Resources", TypeSubscript(Type.Sequence, [Type.str])),
        Argument("Tags", TypeSubscript(Type.Sequence, [TagTypeDef])),
        Argument("DryRun", Type.bool, Type.Ellipsis),
    ),
    return_type=Type.none,
    docstring=(
        "Adds or overwrites only the specified tags for"
        " the specified Amazon EC2 resource or resources."
    ),
)

delete_tags_method = Method(
    name="delete_tags",
    arguments=(
        Argument.self(),
        Argument.kwflag(),
        Argument("Resources", TypeSubscript(Type.Sequence, [Type.str])),
        Argument("Tags", TypeSubscript(Type.Sequence, [TagTypeDef]), Type.Ellipsis),
        Argument("DryRun", Type.bool, Type.Ellipsis),
    ),
    return_type=Type.none,
    docstring="Deletes the specified set of tags from the specified set of resources.",
)

CLIENT_METHODS = (
    create_tags_method.copy(),
    delete_tags_method.copy(),
)
INSTANCE_METHODS = (
    create_tags_method.copy().remove_argument("Resources"),
    delete_tags_method.copy().remove_argument("Resources"),
)
COMMON_METHODS = (create_tags_method.copy().remove_argument("Resources"),)



"""
Type annotations for aiobotocore.retries.standard module.

Copyright 2025 Vlad Emelianov
"""

from typing import Any

from botocore.client import BaseClient
from botocore.retries.standard import (
    OrRetryChecker,
    RetryContext,
    RetryHandler,
    RetryPolicy,
    StandardRetryConditions,
)

def register_retry_handler(client: BaseClient, max_attempts: int = ...) -> AioRetryHandler: ...

class AioRetryHandler(RetryHandler):
    async def needs_retry(self, **kwargs: Any) -> float | None: ...  # type: ignore

class AioRetryPolicy(RetryPolicy):
    async def should_retry(self, context: RetryContext) -> bool: ...

class AioStandardRetryConditions(StandardRetryConditions):
    def __init__(self, max_attempts: int = ...) -> None: ...
    async def is_retryable(self, context: RetryContext) -> bool: ...  # type: ignore

class AioOrRetryChecker(OrRetryChecker):
    async def is_retryable(self, context: RetryContext) -> bool: ...  # type: ignore



"""
Parser that produces `structures.MasterPackage`.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterable

from mypy_boto3_builder.package_data import Boto3StubsPackageData
from mypy_boto3_builder.parsers.fake_service_package import parse_fake_service_package
from mypy_boto3_builder.parsers.resource_loader import ResourceLoader
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.packages.mypy_boto3_package import MypyBoto3Package
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral


def parse_mypy_boto3_package(
    service_names: Iterable[ServiceName],
    version: str,
) -> MypyBoto3Package:
    """
    Parse data for mypy-boto3 package.

    Arguments:
        service_names -- All available service names.
        version -- Package version.

    Returns:
        MypyBoto3Package structure.
    """
    result = MypyBoto3Package(service_names=service_names, service_packages=[], version=version)
    for service_name in result.service_names:
        result.service_packages.append(
            parse_fake_service_package(service_name, Boto3StubsPackageData(), version),
        )

    if service_names:
        result.literals.append(TypeLiteral("ServiceName", [i.boto3_name for i in service_names]))

    resource_service_names = ResourceLoader().get_resource_service_names(service_names)
    if resource_service_names:
        result.literals.append(
            TypeLiteral("ResourceServiceName", [i.boto3_name for i in resource_service_names]),
        )

    return result



"""
Generator for boto3-stubs packages.

Copyright 2024 Vlad Emelianov
"""

from pathlib import Path

from mypy_boto3_builder.constants import (
    StaticStubsPath,
    StaticStubsPullURL,
    TemplatePath,
)
from mypy_boto3_builder.exceptions import AlreadyPublishedError
from mypy_boto3_builder.generators.base_generator import BaseGenerator
from mypy_boto3_builder.package_data import (
    Boto3StubsCustomPackageData,
    Boto3StubsFullPackageData,
    Boto3StubsLitePackageData,
    Boto3StubsPackageData,
)
from mypy_boto3_builder.parsers.parse_wrapper_package import parse_types_boto3_package
from mypy_boto3_builder.postprocessors.botocore import BotocorePostprocessor
from mypy_boto3_builder.structures.package import Package
from mypy_boto3_builder.structures.packages.service_package import ServicePackage
from mypy_boto3_builder.structures.packages.types_boto3_package import TypesBoto3Package


class Boto3Generator(BaseGenerator):
    """
    Generator for boto3-stubs packages.
    """

    service_package_data = Boto3StubsPackageData()
    service_template_path = TemplatePath.types_boto3_service

    def _get_postprocessor(self, service_package: ServicePackage) -> BotocorePostprocessor:
        """
        Get postprocessor for service package.
        """
        return BotocorePostprocessor(service_package, self.main_service_names)

    def _get_static_files_path(self) -> Path:
        return self._get_or_download_static_files_path(
            StaticStubsPath.types_boto3,
            StaticStubsPullURL.types_boto3,
        )

    def generate_stubs(self) -> TypesBoto3Package | None:
        """
        Generate `boto3-stubs` package.
        """
        package_data = Boto3StubsPackageData()
        try:
            version = self._get_package_build_version(package_data.pypi_name)
        except AlreadyPublishedError:
            return None

        self._log_generate(package_data.pypi_name, version)
        package = parse_types_boto3_package(
            service_names=self.main_service_names,
            package_data=package_data,
            version=version,
        )
        package.extras.extend(self._get_wrapper_package_extras(package))
        self.package_writer.write_package(
            package=package,
            template_path=TemplatePath.types_boto3,
            static_files_path=self._get_static_files_path(),
        )
        return package

    def generate_stubs_lite(self) -> Package | None:
        """
        Generate `boto3-stubs-lite` package.
        """
        package_data = Boto3StubsLitePackageData()
        try:
            version = self._get_package_build_version(package_data.pypi_name)
        except AlreadyPublishedError:
            return None

        self._log_generate(package_data.pypi_name, version)
        package = parse_types_boto3_package(
            service_names=self.main_service_names,
            package_data=package_data,
            version=version,
        )
        package.extras.extend(self._get_wrapper_package_extras(package))
        package.set_description(package.get_short_description("Lite type annotations"))
        self.package_writer.write_package(
            package=package,
            template_path=TemplatePath.types_boto3,
            static_files_path=self._get_static_files_path(),
            exclude_template_names=[
                "session.pyi.jinja2",
                "__init__.pyi.jinja2",
            ],
        )
        return package

    def generate_docs(self) -> None:
        """
        Generate service and docs.
        """
        package_data = Boto3StubsPackageData()
        total_str = f"{len(self.service_names)}"

        self.logger.info(
            f"Generating {package_data.pypi_name} package docs",
            tags=package_data.pypi_name,
        )
        package = parse_types_boto3_package(
            service_names=self.service_names,
            package_data=package_data,
            version=self.version,
        )

        self.package_writer.write_docs(
            package=package,
            templates_path=TemplatePath.types_boto3_docs,
        )

        for index, service_name in enumerate(self.service_names):
            current_str = f"{{:0{len(total_str)}}}".format(index + 1)
            package_name = package_data.get_service_package_name(service_name)
            self.logger.info(f"[{current_str}/{total_str}] Generating {package_name} module docs")
            self._process_service_docs(
                service_name=service_name,
                package_data=package_data,
                templates_path=TemplatePath.types_boto3_service_docs,
                version=self.version,
            )

    def generate_full_stubs(self) -> Package | None:
        """
        Generate `boto3-stubs-full` package.
        """
        package_data = Boto3StubsFullPackageData()
        try:
            version = self._get_package_build_version(package_data.pypi_name)
        except AlreadyPublishedError:
            return None

        self._log_generate(package_data.pypi_name, version)
        package = parse_types_boto3_package(
            service_names=self.service_names,
            package_data=package_data,
            version=version,
        )
        package.setup_package_data.clear()
        for service_name in self.service_names:
            package.setup_package_data[package_data.get_service_package_name(service_name)] = (
                "py.typed",
                "*.pyi",
            )
        package.set_description(package.get_short_description("All-in-one type annotations"))
        self.package_writer.write_package(
            package=package,
            template_path=TemplatePath.types_boto3_full,
            static_files_path=None,
        )
        self._generate_full_stubs_services(package)
        self.setup_package_writer.write_package(
            package,
            template_path=TemplatePath.types_boto3_full,
            include_template_names=("setup.py.jinja2",),
        )
        return package

    def generate_custom_stubs(self) -> Package:
        """
        Generate `types-boto3-custom` package.
        """
        package_data = Boto3StubsCustomPackageData()

        self._log_generate(package_data.pypi_name, self.version)
        package = parse_types_boto3_package(
            service_names=self.service_names,
            package_data=package_data,
            version=self.version,
        )
        for service_name in self.service_names:
            package.setup_package_data[package_data.get_service_package_name(service_name)] = (
                "py.typed",
                "*.pyi",
            )
        package.set_description(package.get_short_description("Custom type annotations"))
        self.package_writer.write_package(
            package=package,
            template_path=TemplatePath.types_boto3_custom,
            static_files_path=self._get_static_files_path(),
        )

        self._generate_full_stubs_services(package)
        self.setup_package_writer.write_package(
            package,
            template_path=TemplatePath.types_boto3_custom,
            include_template_names=("setup.py.jinja2",),
        )
        return package



"""
Wrapper for name Union type annotations, like `MyUnion = Union[str, int]`.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Generator, Iterable
from pathlib import Path
from typing import Final, Self

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.exceptions import TypeAnnotationError
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.internal_import_record import InternalImportRecord
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_def_sortable import TypeDefSortable
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.utils.jinja2 import render_jinja2_template


class TypeUnion(TypeSubscript, TypeDefSortable):
    """
    Wrapper for name Union type annotations, like `MyUnion = Union[str, int]`.
    """

    _MIN_CHILDREN: Final = 2

    def __init__(
        self,
        children: Iterable[FakeAnnotation],
        name: str = "",
        *,
        stringify: bool = False,
    ) -> None:
        self.name = name
        self.parent = Type.Union
        self.children: list[FakeAnnotation] = list(children)
        if len(self.children) < self._MIN_CHILDREN:
            raise TypeAnnotationError(f"Union must have at least {self._MIN_CHILDREN} children")
        self._stringify = stringify

    def is_stringified(self) -> bool:
        """
        Whether Union usage should be rendered as a string.
        """
        return self._stringify

    def stringify(self) -> None:
        """
        Render Union usage as a string.
        """
        self._stringify = True

    def render(self) -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        if not self.is_named():
            return " | ".join([i.render() for i in self.children])

        result = self.name

        if self._stringify:
            result = f'"{result}"'

        return result

    def is_named(self) -> bool:
        """
        Check if type annotation is a named type annotation.
        """
        return bool(self.name)

    def __copy__(self) -> Self:
        """
        Create a copy of type annotation wrapper.
        """
        return self.__class__(
            children=list(self.children),
            name=self.name,
            stringify=self._stringify,
        )

    def get_children_types(self) -> set[FakeAnnotation]:
        """
        Extract required type annotations from attributes.
        """
        result: set[FakeAnnotation] = set()
        for child in self.children:
            result.update(child.iterate_types())
        return result

    def get_children_literals(self, processed: Iterable[str] = ()) -> set[TypeLiteral]:
        """
        Extract required TypeLiteral list from attributes.
        """
        result: set[TypeLiteral] = set()
        if self.name in processed:
            return result
        children_types = self.get_children_types()
        for type_annotation in children_types:
            if isinstance(type_annotation, TypeLiteral):
                result.add(type_annotation)
            if isinstance(type_annotation, TypeDefSortable):
                result.update(type_annotation.get_children_literals((self.name, *processed)))
        return result

    def get_definition_import_records(self) -> set[ImportRecord]:
        """
        Get import record required for using Union.
        """
        result: set[ImportRecord] = set()
        if self.is_named():
            result.update(self.parent.get_import_records())
        for child in self.children:
            result.update(child.get_import_records())
        return result

    def _get_import_records(self) -> set[ImportRecord]:
        """
        Get import record required for using type annotation.
        """
        if self.is_named():
            return {InternalImportRecord(ServiceModuleName.type_defs, name=self.name)}

        return self.get_definition_import_records()

    def get_import_records(self) -> set[ImportRecord]:
        """
        Get all import records required for using type annotation.
        """
        return self._get_import_records()

    def iterate_types(self) -> Generator[FakeAnnotation]:
        """
        Extract type annotations from children.
        """
        if self.is_named():
            yield self
            return

        for child in self.children:
            yield from child.iterate_types()

    def is_type_def(self) -> bool:
        """
        Whether type annotation is a TypeDef.
        """
        return self.is_named()

    @property
    def type_hint_annotations(self) -> list[FakeAnnotation]:
        """
        Type annotations list from arguments and return type with internal types.
        """
        return [child for child in self.children if child.get_local_types()]

    def render_definition(self) -> str:
        """
        Render type annotation definition.
        """
        if self.is_named():
            return render_jinja2_template(Path("common/named_union.py.jinja2"), {"type_def": self})

        return self.render()



"""
Path utils.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Generator, Iterable
from pathlib import Path


def print_path(path: Path) -> str:
    """
    Get path as a string relative to current workdir.
    """
    if path.is_absolute():
        cwd = Path.cwd()
        if path == cwd or path.parts <= cwd.parts:
            return path.as_posix()

        try:
            path = path.relative_to(cwd)
        except ValueError:
            return str(path)

    if len(path.parts) == 1:
        return f"./{path.as_posix()}"

    return path.as_posix()


def walk_path(
    parent: Path,
    exclude: Iterable[Path] = (),
    glob_pattern: str = "**/*",
) -> Generator[Path, None, None]:
    """
    Walk files except for `exclude`.

    Yields:
        Existing Path.
    """
    exclude_strs = {Path(i).as_posix() for i in exclude}
    for path in parent.glob(glob_pattern):
        if not path.is_file():
            continue

        if path.as_posix() in exclude_strs:
            continue

        yield path



"""
Type annotations for aioboto3.experimental.async_chalice module.

Copyright 2025 Vlad Emelianov
"""

from typing import Any

from aioboto3.session import Session as Session
from chalice import Chalice  # type: ignore
from chalice.app import RestAPIEventHandler  # type: ignore

class AsyncRestAPIEventHandler(RestAPIEventHandler): ...  # type: ignore

class AsyncChalice(Chalice):  # type: ignore
    aioboto3: Any
    def __init__(
        self, *args: Any, aioboto3_session: Session | None = ..., **kwargs: Any
    ) -> None: ...
    lambda_context: Any
    current_request: Any
    def __call__(self, event: Any, context: Any) -> Any: ...



"""
Version manager for PyPI packages.

Copyright 2024 Vlad Emelianov
"""

import requests

from mypy_boto3_builder.constants import REQUEST_TIMEOUT
from mypy_boto3_builder.exceptions import BuildEnvError
from mypy_boto3_builder.utils.version import bump_postrelease, get_release_version, sort_versions


class PyPIManager:
    """
    Version manager for PyPI packages.

    Arguments:
        package -- PyPI package name
    """

    JSON_URL = "https://pypi.org/pypi/{package}/json"

    def __init__(self, package: str) -> None:
        self.package = package
        self._versions: set[str] | None = None

    @property
    def json_url(self) -> str:
        """
        Package JSON URL on PyPI.
        """
        return self.JSON_URL.format(package=self.package)

    def has_version(self, version: str) -> bool:
        """
        Check if version is already uploaded to PyPI.

        Arguments:
            version -- Target version
        """
        return version in self._get_versions()

    def get_next_version(self, version: str) -> str:
        """
        Get not existing version or closest not existing post-release.

        Arguments:
            version -- Target version
        """
        versions = self._get_versions()
        new_version = version
        while new_version in versions:
            new_version = bump_postrelease(new_version)
        return new_version

    def get_latest_stable_version(self) -> str:
        """
        Get latest stable package version from PyPI.
        """
        versions = self._get_versions()
        sorted_versions = sort_versions(versions)
        if not versions:
            raise BuildEnvError(f"No versions found for {self.package}")

        return get_release_version(sorted_versions[-1])

    def _get_versions(self) -> set[str]:
        if self._versions is not None:
            return self._versions

        response = requests.get(self.json_url, timeout=REQUEST_TIMEOUT)
        if response.status_code == requests.codes.not_found:
            return set()
        if not response.ok:
            raise BuildEnvError(
                f"Cannot retrieve {self.json_url}: {response.status_code} {response.text}",
            ) from None

        data = response.json()

        if "releases" not in data:
            return set()

        version_strs = set(data["releases"].keys())
        self._versions = version_strs
        return self._versions



"""
Boto3 ServiceResource or Resource collection.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Generator

from mypy_boto3_builder.constants import SERVICE_RESOURCE
from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.class_record import ClassRecord
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation


class Collection(ClassRecord):
    """
    Boto3 ServiceResource or Resource collection.
    """

    def __init__(
        self,
        *,
        name: str,
        attribute_name: str,
        parent_name: str,
        service_name: ServiceName,
        type_annotation: FakeAnnotation,
        object_class_name: str,
    ) -> None:
        super().__init__(
            name=name,
            use_alias=True,
            bases=[
                ExternalImport(
                    Import.boto3 + "resources" + "collection",
                    "ResourceCollection",
                )
            ],
        )
        self.service_name = service_name
        self.attribute_name = attribute_name
        self.parent_name = parent_name
        self.type_annotation = type_annotation
        self.object_class_name = object_class_name

    @property
    def boto3_doc_link_parent(self) -> str:
        """
        Link to boto3 docs parent directory.
        """
        parent_name_part = (
            "service-resource" if self.parent_name == SERVICE_RESOURCE else self.parent_name.lower()
        )
        return (
            f"{self.service_name.boto3_doc_link_parent}"
            f"/{parent_name_part}/{self.attribute_name}.html"
        )

    @property
    def boto3_doc_link(self) -> str:
        """
        Link to boto3 docs.
        """
        return (
            f"{self.boto3_doc_link_parent}"
            f"#{self.service_name.class_name}.{self.parent_name}.{self.attribute_name}"
        )

    def iterate_types(self) -> Generator[FakeAnnotation]:
        """
        Iterate over all type annotations.
        """
        yield from super().iterate_types()
        yield from self.type_annotation.iterate_types()



"""
Type annotations for aiobotocore.httpsession module.

Copyright 2025 Vlad Emelianov
"""

from types import TracebackType
from typing import Any, TypeVar

from botocore.endpoint import MAX_POOL_CONNECTIONS as MAX_POOL_CONNECTIONS
from requests.models import Request, Response

_R = TypeVar("_R")

class AIOHTTPSession:
    def __init__(
        self,
        verify: bool = ...,
        proxies: dict[str, str] | None = ...,
        timeout: float | None = ...,
        max_pool_connections: int = ...,
        socket_options: Any | None = ...,
        client_cert: Any | None = ...,
        proxies_config: Any | None = ...,
        connector_args: Any | None = ...,
    ) -> None: ...
    async def __aenter__(self: _R) -> _R: ...
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        tb: TracebackType | None,
    ) -> None: ...
    async def close(self) -> None: ...
    async def send(self, request: Request) -> Response: ...



"""
Type annotations for aiobotocore.stub module.

Copyright 2025 Vlad Emelianov
"""

from typing import Any, Mapping

from botocore.stub import Stubber

class AioStubber(Stubber):
    def add_client_error(
        self,
        method: str,
        service_error_code: str = ...,
        service_message: str = ...,
        http_status_code: int = ...,
        service_error_meta: Mapping[str, Any] | None = ...,
        expected_params: Mapping[str, Any] | None = ...,
        response_meta: Mapping[str, Any] | None = ...,
        modeled_fields: Mapping[str, Any] | None = ...,
    ) -> None: ...



"""
Wrapper for `typing` type annotation.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Mapping
from typing import Final, Self

from mypy_boto3_builder.exceptions import TypeAnnotationError
from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.utils.version import VersionParts


class TypeAnnotation(FakeAnnotation):
    """
    Wrapper for `typing` type annotation.

    Arguments:
        wrapped_type -- Original type annotation as a string.
    """

    # Set of supported type annotations. value is default import module
    _SUPPORTED_TYPES: Final[Mapping[str, ImportRecord]] = {
        "Union": ImportRecord(Import.typing, "Union"),
        "Any": ImportRecord(Import.typing, "Any"),
        "Dict": ImportRecord(
            Import.builtins,
            "dict",
            alias="Dict",
            fallback=ImportRecord(Import.typing, "Dict"),
            min_version=(3, 9),
        ),
        "List": ImportRecord(
            Import.builtins,
            "list",
            alias="List",
            fallback=ImportRecord(Import.typing, "List"),
            min_version=(3, 9),
        ),
        "Set": ImportRecord(
            Import.builtins,
            "set",
            alias="Set",
            fallback=ImportRecord(Import.typing, "Set"),
            min_version=(3, 9),
        ),
        "Type": ImportRecord(
            Import.builtins,
            "type",
            alias="Type",
            fallback=ImportRecord(Import.typing, "Type"),
            min_version=(3, 9),
        ),
        "IO": ImportRecord(Import.typing, "IO"),
        "overload": ImportRecord(Import.typing, "overload"),
        "NoReturn": ImportRecord(Import.typing, "NoReturn"),
        "TypedDict": ImportRecord(
            Import.typing,
            "TypedDict",
            fallback=ImportRecord(Import.typing_extensions, "TypedDict"),
            min_version=(3, 12),
        ),
        "Literal": ImportRecord(
            Import.typing,
            "Literal",
            fallback=ImportRecord(Import.typing_extensions, "Literal"),
            min_version=(3, 12),
        ),
        "Mapping": ImportRecord(
            Import.collections_abc,
            "Mapping",
            fallback=ImportRecord(Import.typing, "Mapping"),
            min_version=(3, 9),
        ),
        "Sequence": ImportRecord(
            Import.collections_abc,
            "Sequence",
            fallback=ImportRecord(Import.typing, "Sequence"),
            min_version=(3, 9),
        ),
        "Callable": ImportRecord(
            Import.collections_abc,
            "Callable",
            fallback=ImportRecord(Import.typing, "Callable"),
            min_version=(3, 9),
        ),
        "Iterator": ImportRecord(
            Import.collections_abc,
            "Iterator",
            fallback=ImportRecord(Import.typing, "Iterator"),
            min_version=(3, 9),
        ),
        "Awaitable": ImportRecord(
            Import.collections_abc,
            "Awaitable",
            fallback=ImportRecord(Import.typing, "Awaitable"),
            min_version=(3, 9),
        ),
        "AsyncIterator": ImportRecord(
            Import.collections_abc,
            "AsyncIterator",
            fallback=ImportRecord(Import.typing, "AsyncIterator"),
            min_version=(3, 9),
        ),
        "NotRequired": ImportRecord(
            Import.typing,
            "NotRequired",
            fallback=ImportRecord(Import.typing_extensions, "NotRequired"),
            min_version=(3, 12),
        ),
        "Unpack": ImportRecord(
            Import.typing,
            "Unpack",
            fallback=ImportRecord(Import.typing_extensions, "Unpack"),
            min_version=(3, 12),
        ),
        "Self": ImportRecord(
            Import.typing,
            "Self",
            fallback=ImportRecord(Import.typing_extensions, "Self"),
            min_version=(3, 12),
        ),
    }

    # Set of fallback type annotations
    _FALLBACK: Final[Mapping[str, tuple[VersionParts, ImportString]]] = {
        "NotRequired": ((3, 12), Import.typing_extensions),
        "TypedDict": ((3, 12), Import.typing_extensions),
        "Literal": ((3, 12), Import.typing_extensions),
        "Unpack": ((3, 12), Import.typing_extensions),
        "Self": ((3, 12), Import.typing_extensions),
        "Mapping": ((3, 9), Import.typing),
        "Sequence": ((3, 9), Import.typing),
        "Callable": ((3, 9), Import.typing),
        "Iterator": ((3, 9), Import.typing),
        "Awaitable": ((3, 9), Import.typing),
        "AsyncIterator": ((3, 9), Import.typing),
        "Dict": ((3, 9), Import.typing),
        "List": ((3, 9), Import.typing),
        "Set": ((3, 9), Import.typing),
        "Type": ((3, 9), Import.typing),
    }

    def __init__(self, wrapped_type: str) -> None:
        if wrapped_type not in self._SUPPORTED_TYPES:
            raise TypeAnnotationError(f"Cannot wrap {wrapped_type}")

        self._wrapped_type: str = wrapped_type

    def render(self) -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        return self.get_import_name()

    def get_import_name(self) -> str:
        """
        Create a safe name for imported annotation.
        """
        return self._wrapped_type

    def _get_import_records(self) -> set[ImportRecord]:
        """
        Create a safe Import Record for annotation.
        """
        name = self.get_import_name()
        import_record = self._SUPPORTED_TYPES[name]
        if import_record.source == Import.builtins and not import_record.fallback:
            return set()
        return {self._SUPPORTED_TYPES[name]}

    def is_dict(self) -> bool:
        """
        Whether annotation is a plain Dict.
        """
        return self._wrapped_type == "Dict"

    def is_list(self) -> bool:
        """
        Whether annotation is a plain List.
        """
        return self._wrapped_type == "List"

    def __copy__(self) -> Self:
        """
        Create a copy of type annotation wrapper.
        """
        return self.__class__(self._wrapped_type)



"""
Methods for boto3 injected methods.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Mapping, Sequence
from typing import Final

from mypy_boto3_builder.constants import CLIENT
from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_maps.service_stub_map import dynamodb, ec2, rds, s3

ClassTypeMap = Mapping[str, Sequence[Method]]
ServiceStubMap = Mapping[ServiceName, ClassTypeMap]

SERVICE_STUB_MAP: Final[ServiceStubMap] = {
    ServiceNameCatalog.ec2: {
        CLIENT: ec2.CLIENT_METHODS,
        "Instance": ec2.INSTANCE_METHODS,
        "DhcpOptions": ec2.COMMON_METHODS,
        "Image": ec2.COMMON_METHODS,
        "InternetGateway": ec2.COMMON_METHODS,
        "NetworkAcl": ec2.COMMON_METHODS,
        "NetworkInterface": ec2.COMMON_METHODS,
        "SecurityGroup": ec2.COMMON_METHODS,
        "Snapshot": ec2.COMMON_METHODS,
        "Volume": ec2.COMMON_METHODS,
        "RouteTable": ec2.COMMON_METHODS,
        "Subnet": ec2.COMMON_METHODS,
        "Vpc": ec2.COMMON_METHODS,
    },
    ServiceNameCatalog.dynamodb: {
        "Table": dynamodb.TABLE_METHODS,
    },
    ServiceNameCatalog.rds: {
        CLIENT: rds.CLIENT_METHODS,
    },
    ServiceNameCatalog.s3: {
        CLIENT: s3.CLIENT_METHODS,
        "Bucket": s3.BUCKET_METHODS,
        "Object": s3.OBJECT_METHODS,
        "ObjectSummary": s3.OBJECT_SUMMARY_METHODS,
    },
}


def get_stub_method_map(service_name: ServiceName, parent: str) -> dict[str, Method]:
    """
    Get boto3 injected methods.
    """
    methods = SERVICE_STUB_MAP.get(service_name, {}).get(parent, [])
    return {method.name: method.copy() for method in methods}



"""
Sortable abstractclass for TypeDefSorter.

Copyright 2024 Vlad Emelianov
"""

from abc import ABC, abstractmethod
from collections.abc import Iterable

from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral


class TypeDefSortable(FakeAnnotation, ABC):
    """
    Sortable abstractclass for TypeDefSorter.
    """

    name: str
    _stringify: bool

    @abstractmethod
    def get_sortable_children(self) -> list["TypeDefSortable"]:
        """
        Extract required sortable TypeDef list from attributes.
        """

    @abstractmethod
    def get_children_literals(self, processed: Iterable[str] = ()) -> set[TypeLiteral]:
        """
        Extract required TypeLiteral list from attributes.
        """

    @property
    @abstractmethod
    def type_hint_annotations(self) -> list[FakeAnnotation]:
        """
        Type annotations list from arguments and return type with internal types.
        """

    @abstractmethod
    def get_definition_import_records(self) -> set[ImportRecord]:
        """
        Get import record required for using TypeAnnotation.
        """

    def is_stringified(self) -> bool:
        """
        Whether TypeDef usage should be rendered as a string.
        """
        return self._stringify

    def stringify(self) -> None:
        """
        Render TypeDef usage as a string.
        """
        self._stringify = True

    @abstractmethod
    def get_children_types(self) -> set[FakeAnnotation]:
        """
        Extract required type annotations from attributes.
        """



"""
Package builder powered by setuptools.

Copyright 2024 Vlad Emelianov
"""

import shutil
import subprocess
import sys
import tarfile
import zipfile
from collections.abc import Sequence
from pathlib import Path
from typing import Literal

from mypy_boto3_builder.enums.output_type import OutputType
from mypy_boto3_builder.exceptions import BuildInternalError
from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.structures.package import Package
from mypy_boto3_builder.utils.path import print_path

_Target = Literal["sdist", "bdist_wheel"]


class PackageBuilder:
    """
    Package builder powered by setuptools.
    """

    def __init__(self, build_path: Path, output_path: Path) -> None:
        self.build_path = build_path
        self.output_path = output_path
        self._logger = get_logger()

    @staticmethod
    def _check_wheel(file_path: Path) -> None:
        try:
            _l = zipfile.ZipFile(file_path).filelist
        except zipfile.BadZipFile:
            raise BuildInternalError(
                f"Failed to build package: {file_path} is not a valid wheel package"
            ) from None

    @staticmethod
    def _check_sdist(file_path: Path) -> None:
        try:
            with tarfile.open(file_path, "r:gz") as f:
                _l = f.getnames()
        except zipfile.BadZipFile:
            raise BuildInternalError(
                f"Failed to build package: {file_path} is not a valid sdist package"
            ) from None

    @classmethod
    def _check_package(cls, file_path: Path) -> bool:
        if file_path.suffix == ".whl":
            cls._check_wheel(file_path)
            return True
        if file_path.suffix == ".gz":
            cls._check_sdist(file_path)
            return True

        return False

    def _build(self, package: Package, targets: Sequence[_Target]) -> None:
        package_path = self.build_path / package.directory_name
        self._logger.debug(f"Building package {print_path(package_path)}")
        dist_path = package_path / "dist"
        if dist_path.exists():
            shutil.rmtree(dist_path)
        try:
            subprocess.check_output(
                [sys.executable, "setup.py", *targets],
                cwd=package_path,
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError as e:
            raise BuildInternalError(f"Failed to build package: {e.output.decode()}") from None

        for file_path in dist_path.iterdir():
            if not file_path.is_file():
                continue

            if not self._check_package(file_path):
                continue

            target_path = self.output_path / file_path.name
            file_path.rename(target_path)
            self._logger.debug(f"Built package {print_path(target_path)}")

    def build(self, package: Package, output_types: Sequence[OutputType]) -> None:
        """
        Build wheel package.
        """
        targets: list[_Target] = []
        if OutputType.wheel in output_types:
            targets.append("bdist_wheel")
        if OutputType.sdist in output_types:
            targets.append("sdist")
        self._build(package, targets)



"""
Profiling entrypoint.

Copyright 2024 Vlad Emelianov
"""

from pathlib import Path

from mypy_boto3_builder.cli_parser import CLINamespace
from mypy_boto3_builder.enums.output_type import OutputType
from mypy_boto3_builder.enums.product import Product
from mypy_boto3_builder.main import run


def main() -> None:
    """
    Profiling entrypoint.
    """
    args = CLINamespace(
        log_level=0,
        output_path=Path("./mypy_boto3_output"),
        service_names=["acm"],
        build_version="1.0.0",
        output_types=[OutputType.package],
        products=[Product.boto3_stubs_services],
        list_services=False,
        partial_overload=False,
        skip_published=False,
        disable_smart_version=True,
        download_static_stubs=False,
    )
    run(args)


if __name__ == "__main__":
    main()



import sys
from unittest.mock import MagicMock, patch

from mypy_boto3_builder.main import get_selected_service_names, main
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.utils.botocore_changelog import BotocoreChangelog


class TestMain:
    def test_get_selected_service_names(self) -> None:
        assert [
            i.name
            for i in get_selected_service_names(
                ["s3", "ec2"],
                [ServiceName("ec2", "EC2"), ServiceName("other", "Other")],
            )
        ] == ["ec2"]
        assert [
            i.name
            for i in get_selected_service_names(
                ["all", "ec2"],
                [ServiceName("ec2", "EC2"), ServiceName("other", "Other")],
            )
        ] == ["ec2", "other"]
        assert get_selected_service_names(["s3", "ec2"], []) == []
        with patch.object(BotocoreChangelog, "fetch_updated") as fetch_updated_mock:
            fetch_updated_mock.return_value = ["ecs"]
            assert [
                i.name
                for i in get_selected_service_names(
                    ["updated", "ec2"],
                    [
                        ServiceName("ec2", "EC2"),
                        ServiceName("ecs", "ECS"),
                        ServiceName("other", "Other"),
                    ],
                )
            ] == ["ec2", "ecs"]

    @patch("mypy_boto3_builder.main.get_available_service_names")
    @patch("mypy_boto3_builder.main.Boto3Generator")
    @patch.object(sys, "argv", ["-o", "/tmp", "-b", "1.2.3.post4"])  # noqa: S108
    def test_main(
        self,
        Boto3GeneratorMock: MagicMock,
        get_available_service_names_mock: MagicMock,
    ) -> None:
        main()
        Boto3GeneratorMock().generate_product.assert_called()



"""
S3 service injected methods.

Copyright 2024 Vlad Emelianov
"""

from botocore.client import BaseClient

from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_constant import TypeConstant
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_maps.named_unions import FileobjTypeDef, PresignedPostConditionsTypeDef
from mypy_boto3_builder.type_maps.typed_dicts import CopySourceTypeDef
from mypy_boto3_builder.utils.type_checks import get_optional

callback_arg = Argument(
    "Callback",
    get_optional(TypeSubscript(Type.Callable, (Type.Ellipsis, Type.Any))),
    Type.Ellipsis,
)

config_arg = Argument(
    "Config",
    get_optional(ExternalImport(Import.boto3 + "s3" + "transfer", "TransferConfig")),
    Type.Ellipsis,
)

extra_args_arg = Argument("ExtraArgs", get_optional(Type.DictStrAny), Type.Ellipsis)

source_client_arg = Argument(
    "SourceClient",
    get_optional(ExternalImport.from_class(BaseClient)),
    Type.Ellipsis,
)

copy_method = Method(
    name="copy",
    arguments=(
        Argument.self(),
        Argument("CopySource", CopySourceTypeDef),
        Argument("Bucket", Type.str),
        Argument("Key", Type.str),
        extra_args_arg,
        callback_arg,
        source_client_arg,
        config_arg,
    ),
    return_type=Type.none,
    docstring="Copy an object from one S3 location to another.",
)

download_file_method = Method(
    name="download_file",
    arguments=(
        Argument.self(),
        Argument("Bucket", Type.str),
        Argument("Key", Type.str),
        Argument("Filename", Type.str),
        extra_args_arg,
        callback_arg,
        config_arg,
    ),
    return_type=Type.none,
    docstring="Download an object from S3 to a file.",
)

download_fileobj_method = Method(
    name="download_fileobj",
    arguments=(
        Argument.self(),
        Argument("Bucket", Type.str),
        Argument("Key", Type.str),
        Argument("Fileobj", FileobjTypeDef),
        extra_args_arg,
        callback_arg,
        config_arg,
    ),
    return_type=Type.none,
    docstring="Download an object from S3 to a file-like object.",
)

generate_presigned_post_method = Method(
    name="generate_presigned_post",
    arguments=(
        Argument.self(),
        Argument("Bucket", Type.str),
        Argument("Key", Type.str),
        Argument("Fields", get_optional(Type.DictStrAny), Type.Ellipsis),
        Argument("Conditions", get_optional(PresignedPostConditionsTypeDef), Type.Ellipsis),
        Argument("ExpiresIn", Type.int, TypeConstant(3600)),
    ),
    return_type=Type.DictStrAny,
    docstring="Generate a presigned URL for POST requests.",
)

upload_file_method = Method(
    name="upload_file",
    arguments=(
        Argument.self(),
        Argument("Filename", Type.str),
        Argument("Bucket", Type.str),
        Argument("Key", Type.str),
        extra_args_arg,
        callback_arg,
        config_arg,
    ),
    return_type=Type.none,
    docstring="Upload a file to S3.",
)

upload_fileobj_method = Method(
    name="upload_fileobj",
    arguments=(
        Argument.self(),
        Argument("Fileobj", FileobjTypeDef),
        Argument("Bucket", Type.str),
        Argument("Key", Type.str),
        extra_args_arg,
        callback_arg,
        config_arg,
    ),
    return_type=Type.none,
    docstring="Upload a file-like object to S3.",
)

bucket_load_method = Method(
    name="load",
    arguments=[Argument.self()],
    return_type=Type.none,
    docstring="Calls s3.Client.list_buckets() to update the attributes of the Bucket resource.",
)

object_summary_load_method = Method(
    name="load",
    arguments=[Argument.self()],
    return_type=Type.none,
    docstring="Calls s3.Client.head_object to update the attributes of the ObjectSummary resource.",
)


CLIENT_METHODS = (
    copy_method.copy(),
    download_file_method.copy(),
    download_fileobj_method.copy(),
    generate_presigned_post_method.copy(),
    upload_file_method.copy(),
    upload_fileobj_method.copy(),
)
BUCKET_METHODS = (
    bucket_load_method.copy(),
    copy_method.copy().remove_argument("Bucket"),
    download_file_method.copy().remove_argument("Bucket"),
    download_fileobj_method.copy().remove_argument("Bucket"),
    upload_file_method.copy().remove_argument("Bucket"),
    upload_fileobj_method.copy().remove_argument("Bucket"),
)
OBJECT_METHODS = (
    copy_method.copy().remove_argument("Bucket", "Key"),
    download_file_method.copy().remove_argument("Bucket", "Key"),
    download_fileobj_method.copy().remove_argument("Bucket", "Key"),
    upload_file_method.copy().remove_argument("Bucket", "Key"),
    upload_fileobj_method.copy().remove_argument("Bucket", "Key"),
)
OBJECT_SUMMARY_METHODS = (object_summary_load_method.copy(),)



"""
DynamoDB service injected methods.

Copyright 2024 Vlad Emelianov
"""

from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.type import Type

batch_writer_method = Method(
    name="batch_writer",
    arguments=(
        Argument.self(),
        Argument(
            "overwrite_by_pkeys",
            Type.wrap_list(Type.str),
            Type.Ellipsis,
        ),
    ),
    return_type=ExternalImport(Import.boto3 + "dynamodb" + "table", "BatchWriter"),
    docstring="Create a batch writer object.",
)

TABLE_METHODS = (batch_writer_method,)



"""
Boto3 client Waiter.

Copyright 2024 Vlad Emelianov
"""

import functools
from typing import Self

from botocore.waiter import Waiter as BotocoreWaiter

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.exceptions import BuildInternalError
from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.class_record import ClassRecord
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral


@functools.total_ordering
class Waiter(ClassRecord):
    """
    Boto3 client Waiter.
    """

    def __init__(
        self,
        name: str,
        waiter_name: str,
        attribute_name: str,
        service_name: ServiceName,
    ) -> None:
        super().__init__(
            name=name,
            bases=[ExternalImport.from_class(BotocoreWaiter)],
        )
        self.waiter_name = waiter_name
        self.attribute_name = attribute_name
        self.service_name = service_name

    def __hash__(self) -> int:
        """
        Hash paginators by name.
        """
        return hash(self.get_sort_key())

    def get_sort_key(self) -> str:
        """
        Sort paginators by name.
        """
        return self.name

    def __eq__(self, other: object) -> bool:
        """
        Whether two annotations are equal.
        """
        if not isinstance(other, Waiter):
            raise BuildInternalError(f"{other} is not Waiter")

        return self.get_sort_key() == other.get_sort_key()

    def __gt__(self, other: Self) -> bool:
        """
        Compare two annotations for sorting.
        """
        return self.get_sort_key() > other.get_sort_key()

    @property
    def boto3_doc_link(self) -> str:
        """
        Link to waiter boto3 docs.
        """
        return (
            f"{self.service_name.boto3_doc_link_parent}"
            f"/waiter/{self.waiter_name}.html"
            f"#{self.service_name.class_name}.Waiter.{self.waiter_name}"
        )

    def get_client_method(self) -> Method:
        """
        Get `get_waiter` method for `Client`.
        """
        return Method(
            name="get_waiter",
            decorators=[Type.overload],
            docstring="Returns an object that can wait for some condition.",
            arguments=(
                Argument.self(),
                Argument("waiter_name", TypeLiteral(f"{self.name}Name", [self.attribute_name])),
            ),
            return_type=ExternalImport(
                source=Import.local(ServiceModuleName.waiter.value),
                name=self.name,
            ),
            type_ignore="override",
        )



"""
Base postprocessor for classes and methods.

Copyright 2024 Vlad Emelianov
"""

from abc import ABC, abstractmethod
from collections.abc import Sequence

from mypy_boto3_builder.exceptions import BuildEnvError
from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.parsers.resource_loader import ResourceLoader
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.structures.packages.service_package import ServicePackage
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_annotations.type_typed_dict import TypedDictAttribute, TypeTypedDict
from mypy_boto3_builder.utils.boto3_utils import get_region_name_literal
from mypy_boto3_builder.utils.strings import textwrap


class BasePostprocessor(ABC):
    """
    Base postprocessor for classes and methods.

    Arguments:
        package -- Service package
        service_names -- Available service names
    """

    def __init__(
        self,
        package: ServicePackage,
        service_names: Sequence[ServiceName],
    ) -> None:
        self.package = package
        self.service_names = service_names
        self.docs_package_name = self.package.data.pypi_name
        self.logger = get_logger()

    def _has_service_resource(self, service_name: ServiceName) -> bool:
        return ResourceLoader().has_service_resource(service_name)

    def generate_docstrings(self) -> None:
        """
        Generate all docstrings.
        """
        self._generate_docstrings_client()
        self._generate_docstrings_paginators()
        self._generate_docstrings_waiters()
        self._generate_docstrings_service_resource()
        self._generate_docstrings_collections()
        self._generate_docstrings_sub_resources()

    @abstractmethod
    def process_package(self) -> None:
        """
        Postprocess built package.
        """

    def _generate_docstrings_client(self) -> None:
        client = self.package.client
        local_doc_link = self.package.get_doc_link("client")
        client.docstring = self._construct_docstring("", client.boto3_doc_link, local_doc_link)
        for method in client.methods:
            boto3_doc_link = (
                f"{self.package.service_name.boto3_doc_link_parent}/client/{method.name}.html"
            )
            if not method.has_boto3_doc_link():
                method.set_boto3_doc_link(boto3_doc_link)
            local_doc_link = self.package.get_doc_link("client", method.name)
            method.docstring = self._construct_docstring(
                method.docstring,
                method.boto3_doc_link,
                local_doc_link,
            )

    def _generate_docstrings_paginators(self) -> None:
        for paginator in self.package.paginators:
            local_doc_link = self.package.get_doc_link("paginators", paginator.name)
            paginator.docstring = self._construct_docstring(
                "",
                paginator.boto3_doc_link,
                local_doc_link,
            )
            for method in paginator.methods:
                boto3_doc_link = f"{paginator.boto3_doc_link}.{method.name}"
                method.set_boto3_doc_link(boto3_doc_link)
                method.docstring = self._construct_docstring(
                    method.docstring,
                    method.boto3_doc_link,
                    local_doc_link,
                )

    def _generate_docstrings_waiters(self) -> None:
        for waiter in self.package.waiters:
            local_doc_link = self.package.get_doc_link("waiters", waiter.name)
            waiter.docstring = self._construct_docstring("", waiter.boto3_doc_link, local_doc_link)
            for method in waiter.methods:
                boto3_doc_link = f"{waiter.boto3_doc_link}.{method.name}"
                method.set_boto3_doc_link(boto3_doc_link)
                method.docstring = self._construct_docstring(
                    method.docstring,
                    method.boto3_doc_link,
                    local_doc_link,
                )

    def _generate_docstrings_service_resource(self) -> None:
        if not self.package.service_resource:
            return

        service_resource = self.package.service_resource
        local_doc_link = self.package.get_doc_link("service_resource")
        service_resource.docstring = self._construct_docstring(
            "",
            service_resource.boto3_doc_link,
            local_doc_link,
        )
        for method in service_resource.methods:
            boto3_doc_link = (
                f"{service_resource.service_name.boto3_doc_link_parent}"
                f"/service-resource/{method.name}.html"
            )
            method.set_boto3_doc_link(boto3_doc_link)
            local_doc_link = self.package.get_doc_link(
                "service_resource",
                service_resource.name,
                f"{method.name} method",
            )
            method.docstring = self._construct_docstring(
                method.docstring,
                method.boto3_doc_link,
                local_doc_link,
            )

    def _generate_docstrings_collections(self) -> None:
        if not self.package.service_resource:
            return

        for collection in self.package.service_resource.collections:
            local_doc_link = self.package.get_doc_link("service_resource", collection.name)
            collection.docstring = self._construct_docstring(
                "",
                collection.boto3_doc_link,
                local_doc_link,
            )
            for method in collection.methods:
                # FIXME: potentially links will be changed to the same as paginators/waiters use
                if not method.has_boto3_doc_link():
                    method.set_boto3_doc_link(f"{collection.boto3_doc_link_parent}#{method.name}")
                method.docstring = self._construct_docstring(
                    method.docstring,
                    method.boto3_doc_link,
                    local_doc_link,
                )

    def _generate_docstrings_sub_resources(self) -> None:
        if not self.package.service_resource:
            return

        for sub_resource in self.package.service_resource.sub_resources:
            local_doc_link = self.package.get_doc_link("service_resource", sub_resource.name)
            sub_resource.docstring = self._construct_docstring(
                "",
                sub_resource.boto3_doc_link,
                local_doc_link,
            )
            for method in sub_resource.methods:
                boto3_doc_link = (
                    f"{sub_resource.service_name.boto3_doc_link_parent}"
                    f"/{sub_resource.name.lower()}"
                    f"/{method.name}.html"
                )
                method.set_boto3_doc_link(boto3_doc_link)
                local_doc_link = self.package.get_doc_link(
                    "service_resource",
                    sub_resource.name,
                    f"{method.name} method",
                )
                method.docstring = self._construct_docstring(
                    method.docstring,
                    method.boto3_doc_link,
                    local_doc_link,
                )
            for collection in sub_resource.collections:
                local_doc_link = self.package.get_doc_link(
                    "service_resource",
                    sub_resource.name,
                    collection.attribute_name,
                )
                collection.docstring = self._construct_docstring(
                    "",
                    collection.boto3_doc_link,
                    local_doc_link,
                )
                for method in collection.methods:
                    if not method.has_boto3_doc_link():
                        method.set_boto3_doc_link(
                            f"{collection.boto3_doc_link_parent}#{method.name}",
                        )
                    method.docstring = self._construct_docstring(
                        method.docstring,
                        method.boto3_doc_link,
                        local_doc_link,
                    )

    def _construct_docstring(self, docstring: str, boto3_doc_link: str, local_doc_link: str) -> str:
        docstring_part = f"{textwrap(docstring)}\n\n" if docstring else ""
        return (
            f"{docstring_part}"
            f"[Show boto3 documentation]({boto3_doc_link})\n"
            f"[Show {self.docs_package_name} documentation]({local_doc_link})"
        )

    def extend_literals(self) -> None:
        """
        Add extra literals.

        - `<Class>ServiceName`
        - `ServiceName`
        - `ResourceServiceName`
        - `PaginatorName`
        - `WaiterName`
        - `RegionName`
        """
        self.package.literals.append(
            TypeLiteral(
                f"{self.package.service_name.class_name}ServiceName",
                [self.package.service_name.boto3_name],
            ),
        )
        if self.service_names:
            self.package.literals.append(
                TypeLiteral("ServiceName", [i.boto3_name for i in self.service_names]),
            )
        resource_service_names = [i for i in self.service_names if self._has_service_resource(i)]
        if resource_service_names:
            self.package.literals.append(
                TypeLiteral("ResourceServiceName", [i.boto3_name for i in resource_service_names]),
            )

        paginator_names = [paginator.operation_name for paginator in self.package.paginators]
        if paginator_names:
            self.package.literals.append(TypeLiteral("PaginatorName", paginator_names))

        waiter_names = [waiter.attribute_name for waiter in self.package.waiters]
        if waiter_names:
            self.package.literals.append(TypeLiteral("WaiterName", waiter_names))

        region_name_literal = get_region_name_literal([self.package.service_name])
        if region_name_literal:
            self.package.literals.append(region_name_literal)

    def _replace_typed_dict_with_dict(
        self,
        attribute: TypedDictAttribute,
        reference: TypeTypedDict,
        parent: TypeTypedDict,
    ) -> None:
        replacement = Type.DictStrAny
        if attribute.type_annotation is reference:
            attribute.type_annotation = replacement
            self.logger.debug(
                f"Replaced {reference.name} with {replacement.render()} in"
                f" {parent.name}.{attribute.name}",
            )
            return

        if isinstance(attribute.type_annotation, TypeSubscript):
            attribute.type_annotation.replace_child(reference, Type.DictStrAny)
            self.logger.debug(
                f"Deep replaced {reference.name} with {replacement.render()} in"
                f" {parent.name}.{attribute.name}",
            )
            return

        raise BuildEnvError(
            f"Cannot replace child {reference.name} in {parent.name}.{attribute.name}",
        )

    def _replace_typed_dict_references(
        self,
        typed_dict: TypeTypedDict,
        reference: TypeTypedDict,
        depth: int,
    ) -> None:
        """
        Replace self references with `Dict[str, Any]` to avoid circular dependencies.
        """
        if depth <= 0:
            return
        next_depth = depth - 1

        for child in typed_dict.iterate_children():
            for child_typed_dict in child.type_annotation.iterate_types():
                if not isinstance(child_typed_dict, TypeTypedDict):
                    continue

                if child_typed_dict is reference:
                    self._replace_typed_dict_with_dict(child, reference, typed_dict)
                if next_depth > 0:
                    self._replace_typed_dict_references(child_typed_dict, typed_dict, next_depth)

    def replace_self_ref_typed_dicts(self) -> None:
        """
        Remove self-references from TypedDicts.
        """
        for type_def in self.package.type_defs:
            if isinstance(type_def, TypeTypedDict):
                self._replace_typed_dict_references(type_def, type_def, 2)



"""
Type annotations for aiobotocore.utils module.

Copyright 2025 Vlad Emelianov
"""

from typing import Any, Mapping

from aiobotocore.credentials import AioCredentials
from botocore.utils import DEFAULT_METADATA_SERVICE_TIMEOUT as DEFAULT_METADATA_SERVICE_TIMEOUT
from botocore.utils import METADATA_BASE_URL as METADATA_BASE_URL
from botocore.utils import (
    ContainerMetadataFetcher,
    IdentityCache,
    IMDSFetcher,
    IMDSRegionProvider,
    InstanceMetadataFetcher,
    InstanceMetadataRegionFetcher,
    S3ExpressIdentityCache,
    S3ExpressIdentityResolver,
    S3RegionRedirector,
    S3RegionRedirectorv2,
)
from requests.models import Response

logger: Any
RETRYABLE_HTTP_ERRORS: Any

class AioIMDSFetcher(IMDSFetcher):
    def __init__(
        self,
        timeout: int = ...,
        num_attempts: int = ...,
        base_url: str = ...,
        env: str | None = ...,
        user_agent: str | None = ...,
        config: Any | None = ...,
        session: Any | None = ...,
    ) -> None: ...

class AioInstanceMetadataFetcher(AioIMDSFetcher, InstanceMetadataFetcher):
    async def retrieve_iam_role_credentials(self) -> dict[str, Any]: ...  # type: ignore[override]

class AioIMDSRegionProvider(IMDSRegionProvider):
    async def provide(self) -> str: ...

class AioInstanceMetadataRegionFetcher(AioIMDSFetcher, InstanceMetadataRegionFetcher):
    async def retrieve_region(self) -> str | None: ...  # type: ignore[override]

class AioIdentityCache(IdentityCache):
    async def get_credentials(self, **kwargs: Any) -> AioCredentials: ...  # type: ignore[override]

class AioS3ExpressIdentityCache(AioIdentityCache, S3ExpressIdentityCache):
    async def get_credentials(self, bucket: str) -> AioCredentials: ...  # type: ignore[override]

class AioS3ExpressIdentityResolver(S3ExpressIdentityResolver): ...

class AioS3RegionRedirectorv2(S3RegionRedirectorv2):
    async def redirect_from_error(
        self,
        request_dict: Mapping[str, Any],
        response: Response,
        operation: Any,
        **kwargs: Any,
    ) -> None: ...
    async def get_bucket_region(self, bucket: Any, response: Response) -> Any: ...

class AioS3RegionRedirector(S3RegionRedirector):
    async def redirect_from_error(
        self, request_dict: Mapping[str, Any], response: Response, operation: Any, **kwargs: Any
    ) -> int | None: ...
    async def get_bucket_region(self, bucket: str, response: Response) -> str: ...

class AioContainerMetadataFetcher(ContainerMetadataFetcher):
    def __init__(self, session: Any | None = ..., sleep: Any = ...) -> None: ...
    async def retrieve_full_uri(self, full_url: str, headers: Any | None = ...) -> str: ...
    async def retrieve_uri(self, relative_uri: str) -> Any: ...



"""
Wrapper for ImportString usage.

Copyright 2024 Vlad Emelianov
"""

from typing import Final

from mypy_boto3_builder.exceptions import StructureError
from mypy_boto3_builder.import_helpers.import_parent import ImportParent
from mypy_boto3_builder.import_helpers.import_string import ImportString


class Import:
    """
    Wrapper for ImportString usage.
    """

    future: Final = ImportString(ImportParent.future.value)
    builtins: Final = ImportString(ImportParent.builtins.value)
    boto3: Final = ImportString(ImportParent.boto3.value)
    botocore: Final = ImportString(ImportParent.botocore.value)
    typing: Final = ImportString(ImportParent.typing.value)
    awscrt: Final = ImportString(ImportParent.awscrt.value)
    s3transfer: Final = ImportString(ImportParent.s3transfer.value)
    aiobotocore: Final = ImportString(ImportParent.aiobotocore.value)
    aioboto3: Final = ImportString(ImportParent.aioboto3.value)
    typing_extensions: Final = ImportString(ImportParent.typing_extensions.value)
    types: Final = ImportString(ImportParent.types.value)
    sys: Final = ImportString(ImportParent.sys.value)
    collections_abc: Final = ImportString(ImportParent.collections.value, "abc")

    @classmethod
    def local(cls, module_name: str) -> ImportString:
        """
        Create local import string.
        """
        if not module_name:
            raise StructureError("Module name cannot be empty for local import")
        return ImportString("", module_name)

    @classmethod
    def from_str(cls, import_string: str) -> ImportString:
        """
        Create from string.
        """
        return cls.from_parts(*import_string.split("."))

    @classmethod
    def from_parts(cls, parent: str, *parts: str) -> ImportString:
        """
        Create from string.
        """
        if not parent:
            raise StructureError("Parent cannot be empty")
        return ImportString(parent, *parts)



"""
Alias map fixes added by botocore for documentation build.

# https://github.com/boto/botocore/blob/develop/botocore/handlers.py#L773
# https://github.com/boto/botocore/blob/develop/botocore/handlers.py#L1055

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Mapping
from typing import Final

from mypy_boto3_builder.constants import ALL
from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.utils.lookup_dict import LookupDict

# Mapping to rename for botocore method arguments.
# ServiceName -> Operation name -> Argument name -> alias name
# DELETE means that argument has to be deleted.
ARGUMENT_ALIASES: Final[Mapping[ServiceName, Mapping[str, Mapping[str, str]]]] = {
    ServiceNameCatalog.cloudsearchdomain: {"Search": {"return": "returnFields"}},
    ServiceNameCatalog.logs: {"CreateExportTask": {"from": "fromTime"}},
    ServiceNameCatalog.ec2: {ALL: {"Filter": "Filters"}},
}

_LOOKUP: LookupDict[str] = LookupDict(
    {ServiceNameCatalog.to_str(k): v for k, v in ARGUMENT_ALIASES.items()},
)


def get_argument_alias(
    service_name: ServiceName,
    operation_name: str,
    argument_name: str,
) -> str:
    """
    Get argument alias for operation.

    Args:
        service_name -- Service name
        operation_name -- Resource operation name
        argument_name -- Argument name

    Returns:
        Argument alias name or original name.
    """
    lookup_argument_name = _LOOKUP.get(service_name.name, operation_name, argument_name)
    return lookup_argument_name or argument_name



"""
String to type annotation map to replace overriden botocore literals.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Mapping
from typing import Final

from mypy_boto3_builder.constants import ALL
from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.utils.lookup_dict import LookupDict

# Mapping reresenting Required/NotRequired keys of output botocore shapes.
# ServiceName -> TypedDict name -> Argument name -> is_required (True/False)
# False means that argument should be marked as NotRequired.
# True means that argument should be marked as Required.
# None or missing value means that argument should be marked as is_required by default.
REQUIRED_ATTRIBUTE_MAP: Final[Mapping[ServiceName, Mapping[str, Mapping[str, bool]]]] = {
    ServiceNameCatalog.all: {
        ALL: {
            "NextToken": False,
            "nextToken": False,
            "Contents": False,
            "Item": False,
            "CommonPrefixes": False,
        },
    },
    ServiceNameCatalog.dynamodb: {
        ALL: {
            "LastEvaluatedKey": False,
        },
    },
    ServiceNameCatalog.stepfunctions: {
        "DescribeExecutionOutputTypeDef": {
            "stopDate": False,
            "output": False,
            "outputDetails": False,
            "error": False,
            "cause": False,
            "traceHeader": False,
            "redriveStatusReason": False,
            "mapRunArn": False,
            "redriveDate": False,
            "stateMachineAliasArn": False,
            "stateMachineVersionArn": False,
        },
    },
}

_LOOKUP: LookupDict[bool] = LookupDict(
    {ServiceNameCatalog.to_str(k): v for k, v in REQUIRED_ATTRIBUTE_MAP.items()},
)


def is_required(service_name: ServiceName, typed_dict_name: str, argument_name: str) -> bool:
    """
    Check if output shape argument should be marked as NotRequired.

    Arguments:
        service_name -- Service name.
        typed_dict_name -- Target TypedDict name.
        attribute_name -- Target attribute name.

    Returns:
        Literal children or None.
    """
    is_required = _LOOKUP.get(service_name.name, typed_dict_name, argument_name)
    return True if is_required is None else is_required



"""
Import replacement map for aio libraries.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Mapping
from typing import Final

from botocore.client import BaseClient
from botocore.config import Config
from botocore.eventstream import EventStream
from botocore.paginate import PageIterator, Paginator
from botocore.response import StreamingBody
from botocore.waiter import Waiter

from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.type_annotations.external_import import ExternalImport

AIO_IMPORT_MAP: Final[Mapping[ExternalImport, ExternalImport]] = {
    ExternalImport.from_class(StreamingBody): ExternalImport(
        Import.aiobotocore + "response", "StreamingBody"
    ),
    ExternalImport.from_class(EventStream): ExternalImport(
        Import.aiobotocore + "eventstream", "AioEventStream"
    ),
    ExternalImport.from_class(Config): ExternalImport(Import.aiobotocore + "config", "AioConfig"),
    ExternalImport.from_class(Waiter): ExternalImport(Import.aiobotocore + "waiter", "AIOWaiter"),
    ExternalImport.from_class(Paginator): ExternalImport(
        Import.aiobotocore + "paginate", "AioPaginator"
    ),
    ExternalImport.from_class(PageIterator): ExternalImport(
        Import.aiobotocore + "paginate", "AioPageIterator"
    ),
    ExternalImport.from_class(BaseClient): ExternalImport(
        Import.aiobotocore + "client", "AioBaseClient"
    ),
    ExternalImport(Import.boto3 + "resources" + "base", "ServiceResource"): ExternalImport(
        Import.aioboto3 + "resources" + "base", "AIOBoto3ServiceResource"
    ),
    ExternalImport(Import.boto3 + "resources" + "collection", "ResourceCollection"): ExternalImport(
        Import.aioboto3 + "resources" + "collection", "AIOResourceCollection"
    ),
    ExternalImport(Import.boto3 + "dynamodb" + "table", "BatchWriter"): ExternalImport(
        Import.aioboto3 + "dynamodb" + "table", "BatchWriter"
    ),
    ExternalImport(Import.boto3 + "dynamodb" + "table", "TableResource"): ExternalImport(
        Import.aioboto3 + "dynamodb" + "table", "CustomTableResource"
    ),
}


def replace_import_with_aio(external_import: ExternalImport) -> None:
    """
    Replace botocore/boto3 import with aioboto3/aiobotocore.
    """
    if external_import not in AIO_IMPORT_MAP:
        return
    new_external_import = AIO_IMPORT_MAP[external_import]
    external_import.copy_from(new_external_import)



"""
Choice for select.

Copyright 2024 Vlad Emelianov
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from questionary.prompts.common import Choice as QuestionaryChoice

from mypy_boto3_builder.chat.text_style import TextStyle
from mypy_boto3_builder.chat.utils import as_message, as_string

if TYPE_CHECKING:
    from collections.abc import Sequence

    from mypy_boto3_builder.chat.type_defs import Message


class Choice(QuestionaryChoice):
    """
    Choice for select.
    """

    def __init__(
        self,
        title: Message | str,
        *,
        key: str | None = None,
        text: Message | str | None = None,
        aliases: Sequence[str] = (),
        shortcut_key: str | None = None,
        selected_suffix: Message | str | None = None,
        disabled: bool = False,
    ) -> None:
        self.raw_title = title
        self.selected_suffix = TextStyle.hilight.apply(selected_suffix or " ✓")
        choice_title = TextStyle.text.stylize(title)
        self.key = key or as_string(choice_title) or "(empty)"
        super().__init__(
            title=choice_title,
            value=self.key,
            shortcut_key=shortcut_key or False,
            disabled="✓" if disabled else None,
        )
        self.text: Message | str = self.key if text is None else text
        self.aliases = aliases or (self.key,)
        self._is_selected = False
        self.select_index = 0

    def __hash__(self) -> int:
        """
        Hashable by key.
        """
        return hash(self.value)

    @classmethod
    def create(cls, choice: Choice | str) -> Choice:
        """
        Create choice from string.
        """
        return cls(choice) if isinstance(choice, str) else choice

    @property
    def tag(self) -> Message:
        """
        Previous answers.
        """
        return TextStyle.tag.apply(self.text)

    @property
    def answer(self) -> Message:
        """
        Current answer.
        """
        return TextStyle.hilight.apply(self.text)

    @property
    def is_selected(self) -> bool:
        """
        Whether choice is selected.
        """
        return self._is_selected

    def select(self, index: int = 0) -> None:
        """
        Select choice.
        """
        self.title = TextStyle.dim.stylize(
            (
                *as_message(self.raw_title),
                *self.selected_suffix,
            )
        )
        self.select_index = index
        self._is_selected = True

    def deselect(self) -> None:
        """
        Deselect choice.
        """
        self.title = TextStyle.text.stylize(self.raw_title)
        self.select_index = 0
        self._is_selected = False



"""
Select prompt based on questionary.path.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterator, Sequence
from contextlib import contextmanager

import questionary
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from prompt_toolkit.layout import ConditionalContainer, Window
from questionary.prompts.common import InquirerControl

from mypy_boto3_builder.chat.choice import Choice
from mypy_boto3_builder.chat.prompts.base_prompt import BasePrompt
from mypy_boto3_builder.chat.text_style import TextStyle
from mypy_boto3_builder.chat.type_defs import Message, MessageToken
from mypy_boto3_builder.chat.utils import as_message


class SelectPrompt(BasePrompt[list[Choice]]):
    """
    Select prompt based on questionary.path.
    """

    PROMPT_SESSION_MODULE = "questionary.prompts.common.PromptSession"
    HELP_MESSAGE = (
        "Select item using ",
        TextStyle.tag.wrap("arrow keys"),
        ", then press ",
        TextStyle.tag.wrap("Enter"),
        " to continue.",
    )

    def __init__(
        self,
        message: Message | str,
        choices: Sequence[Choice | str],
        *,
        default: Choice | str | None = None,
        message_end: Message | str = "",
        instruction: Message | str = "",
    ) -> None:
        super().__init__()
        self.message = as_message(message)
        self.choices = tuple(Choice.create(choice) for choice in choices)
        self.message_end = as_message(message_end)
        self.instruction = as_message(instruction) or self.get_help_message()
        self.default = default
        self._inquirer_control: InquirerControl | None = None

    @property
    def inquirer_control(self) -> InquirerControl:
        """
        InquirerControl of a currently running application.
        """
        if self._inquirer_control is None:
            raise ValueError("InquirerControl is not set")
        return self._inquirer_control

    def _get_pointed_at(self) -> Choice:
        choice = self.inquirer_control.get_pointed_at()
        if not isinstance(choice, Choice):
            raise TypeError(f"Expected Choice, got {choice}")
        return choice

    def _patch_answer(self) -> None:
        enter_key_binding = self._get_enter_key_binding()

        def set_answer(event: KeyPressEvent) -> None:
            selected_choice = self._get_pointed_at()
            result = [selected_choice]
            self._set_result(result)
            self.inquirer_control.is_answered = True
            event.app.exit(result=result)

        enter_key_binding.handler = set_answer

    @contextmanager
    def _patch_session(self) -> Iterator[None]:
        with super()._patch_session():
            # patch shortcut keys because questionary.select check is incorrect
            original_shortcut_keys = list(InquirerControl.SHORTCUT_KEYS)
            InquirerControl.SHORTCUT_KEYS = original_shortcut_keys * 100
            yield
            InquirerControl.SHORTCUT_KEYS = original_shortcut_keys
            self._inquirer_control = self._get_inquirer_control()

    def _get_prompt_tokens(self) -> Message:
        selected_choices = sorted(
            filter(lambda x: x.is_selected and not x.disabled, self.choices),
            key=lambda x: x.select_index,
        )

        tokens: list[MessageToken] = []
        if not self.inquirer_control.is_answered:
            tokens.extend(TextStyle.dim.apply(self.instruction))

        tokens.extend(as_message(self.message))

        selected_choice = self._get_pointed_at()
        selected_tokens = (
            *(i.tag for i in selected_choices),
            *(
                (selected_choice.answer,)
                if selected_choice.text and not selected_choice.is_selected
                else ()
            ),
        )
        tokens.extend(self._format_selected_tokens(selected_tokens))
        tokens.extend(as_message(self.message_end))

        return tokens

    def _format_selected_tokens(self, raw_selected: Sequence[Message]) -> Message:
        selected = [item for item in raw_selected if item]
        if not selected:
            return []

        if len(selected) == 1:
            return selected[0]

        if len(selected) == 1 + 1:
            return [*selected[0], " and ", *selected[1]]

        result: list[MessageToken] = []
        for item in selected[:-1]:
            result.extend((*item, ", "))
        result.extend(("and ", *selected[-1]))
        return result

    def _create_question(self) -> questionary.Question:
        """
        Create an underlying question.
        """
        return questionary.select(
            message="",
            choices=self.choices,
            default=self.default,
            style=self.STYLE,
            use_shortcuts=True,
            use_jk_keys=False,
        )

    def _get_inquirer_control(self) -> InquirerControl:
        window_container = self.question.application.layout.container.get_children()[1]
        if not isinstance(window_container, ConditionalContainer):
            raise TypeError("Unexpected layout")

        window = window_container.content
        if not isinstance(window, Window):
            raise TypeError("Unexpected layout")

        result = window.content
        if not isinstance(result, InquirerControl):
            raise TypeError("Unexpected layout")
        return result



from collections.abc import Iterator
from unittest.mock import MagicMock, patch

import pytest

from mypy_boto3_builder.cli_parser import get_absolute_path, parse_args


class TestCLIParser:
    @pytest.fixture(autouse=True)
    def _patch_argparse(self) -> Iterator[None]:
        with patch("mypy_boto3_builder.cli_parser.argparse"):
            yield

    def test_parse_args(self) -> None:
        result = parse_args([])
        assert result

    @patch("mypy_boto3_builder.cli_parser.Path")
    def test_get_absolute_path(self, PathMock: MagicMock) -> None:
        result = get_absolute_path("test/output")
        PathMock.assert_called_with("test/output")
        assert result == PathMock().absolute()



"""
Product type choice for CLI.

Copyright 2024 Vlad Emelianov
"""

from enum import Enum


class ProductType(Enum):
    """
    Product type choice for CLI.
    """

    stubs = "stubs"
    stubs_lite = "stubs_lite"
    service_stubs = "service_stubs"
    docs = "docs"
    full = "full"
    custom = "custom"
